# encoding: utf-8

from __future__ import unicode_literals

import six
from django.db.models import Manager
from django.db.models.query import QuerySet

from .compat import (ANNOTATION_SELECT_CACHE_NAME, ANNOTATION_TO_AGGREGATE_ATTRIBUTES_MAP, chain_query, chain_queryset,
                     LOOKUP_SEP, ModelIterable, ValuesQuerySet)
from .exceptions import QueryablePropertyDoesNotExist, QueryablePropertyError
from .query import QueryablePropertiesQueryMixin, QueryablePropertyReference
from .utils import get_queryable_property, InjectableMixin


class QueryablePropertiesIterable(InjectableMixin):
    """
    An iterable that yields the actual results of a queryset while correctly
    processing columns of queryable properties. It is closely related to
    Django's BaseIterable and will be used as a mixin for its subclasses in all
    (recent) Django versions that have it. In all other (older) versions, this
    class will be used as a standalone iterable instead.
    """

    def __init__(self, queryset, *args, **kwargs):
        """
        Initialize a new iterable for the given queryset. If an iterable is
        given it will be used to retrieve the model instances before applying
        queryable properties logic (standalone usage for older Django
        versions). Otherwise, the __iter__ implementation of the base class
        is used to get the model instances (usage as mixin).

        :param QuerySet queryset: The queryset to perform the database query
                                  for.
        :param collections.Iterable iterable: The optional iterable to use for
                                              standalone usage.
        :param args: Positional arguments to pass through to the base class
                     initialization when used as a mixin.
        :param kwargs: Keyword arguments to pass through to the base class
                       initialization when used as a mixin.
        :keyword collections.Iterable iterable: The optional iterable to use
                                                for standalone usage.
        """
        self.queryset = queryset
        # Only perform the super call if the class is used as a mixin
        if self.__class__.__bases__ != (InjectableMixin,):
            super(QueryablePropertiesIterable, self).__init__(queryset, *args, **kwargs)
        self.iterable = kwargs.get('iterable') or super(QueryablePropertiesIterable, self).__iter__()
        self.yields_model_instances = ((ModelIterable is not None and isinstance(self, ModelIterable)) or
                                       (ValuesQuerySet is not None and not isinstance(self.queryset, ValuesQuerySet)))

    def __iter__(self):
        """
        Yield the model objects for the queryset associated with this iterator
        with their correctly processed selected queryable properties.

        :return: A generator that yields the model objects.
        """
        original_query = self.queryset.query
        try:
            self.queryset.query = chain_query(original_query)
            final_aliases = self._setup_queryable_properties()

            for obj in self.iterable:
                if self.yields_model_instances:
                    # Retrieve the annotation values from each renamed
                    # attribute and use it to populate the cache for the
                    # corresponding queryable property on each object while
                    # removing the weird, renamed attributes.
                    for changed_name, property_ref in six.iteritems(final_aliases):
                        value = getattr(obj, changed_name)
                        delattr(obj, changed_name)
                        if property_ref:
                            property_ref.property._set_cached_value(obj, value)
                yield obj
        finally:
            self.queryset.query = original_query

    def _setup_queryable_properties(self):
        """
        Perform the required setup to correctly process queryable property
        values.

        Change the internal aliases of the annotations that belong to queryable
        properties in the query of the associated queryset to something unique
        and return a dictionary mapping the queryable properties to the changed
        aliases. This is necessary to allow Django to populate the annotation
        attributes on the resulting model instances, which would otherwise call
        the setter of the queryable properties. This way, Django can populate
        attributes with different names and avoid using the setter methods.

        Also make sure that ordering by queryable properties works in older
        Django versions.

        :return: A dictionary mapping the final aliases for queryable
                 properties to the corresponding references to be able to
                 retrieve the values from the DB and apply them to the correct
                 property. The property reference may be None, indicating that
                 the retrieved value should be discarded.
        :rtype: dict[str, QueryablePropertyReference | None]
        """
        query = self.queryset.query
        final_aliases = {}
        select = dict(query.annotation_select)

        for property_ref in query._queryable_property_annotations:
            annotation_name = property_ref.full_path

            # Older Django versions don't work with the annotation select dict
            # when it comes to ordering, so queryable property annotations used
            # for ordering need special treatment.
            order_by_occurrences = []
            if ANNOTATION_TO_AGGREGATE_ATTRIBUTES_MAP:  # pragma: no cover
                order_by_occurrences = [index for index, field_name in enumerate(query.order_by)
                                        if field_name == annotation_name or field_name[1:] == annotation_name]
                if order_by_occurrences and annotation_name not in select and annotation_name in query.annotations:
                    select[annotation_name] = query.annotations[annotation_name]
                    final_aliases[annotation_name] = None

            if not self.yields_model_instances or annotation_name not in select:
                # The queryable property annotation does not require selection
                # or no renaming needs to occur since the queryset doesn't
                # yield model instances.
                continue

            # Suffix the original annotation name with the lookup separator to
            # create a non-clashing name: both model field an queryable
            # property names are not allowed to contain the separator and a
            # relation path ending with the separator would be invalid as well.
            changed_name = ''.join((annotation_name, LOOKUP_SEP))
            final_aliases[changed_name] = final_aliases.pop(annotation_name, property_ref)
            select[changed_name] = select.pop(annotation_name)
            for index in order_by_occurrences:  # pragma: no cover
                # Apply the changed names to the ORDER BY clause.
                query.order_by[index] = query.order_by[index].replace(annotation_name, changed_name)

        # Patch the correct select property on the query with the new names,
        # since this property is used by the SQL compiler to build the actual
        # SQL query (which is where the changed names should be used).
        setattr(query, ANNOTATION_SELECT_CACHE_NAME, select)
        return final_aliases


class QueryablePropertiesQuerySetMixin(InjectableMixin):
    """
    A mixin for Django's :class:`django.db.models.QuerySet` objects that allows
    to use queryable properties in filters, annotations and update queries.
    """

    def __init__(self, *args, **kwargs):
        super(QueryablePropertiesQuerySetMixin, self).__init__(*args, **kwargs)
        # To work correctly, a query using the QueryablePropertiesQueryMixin is
        # required. If the current query is not using the mixin already, it
        # will be dynamically injected into the query. That way, other Django
        # extensions using custom query objects are also supported.
        if not isinstance(self.query, QueryablePropertiesQueryMixin):
            self.query = chain_query(self.query)
            class_name = 'QueryableProperties' + self.query.__class__.__name__
            QueryablePropertiesQueryMixin.inject_into_object(self.query, class_name)

    @property
    def _iterable_class(self):
        # Override the regular _iterable_class attribute of recent Django
        # versions with a property that also stores the value in the instance
        # dict, but automatically mixes the QueryablePropertiesModelIterable
        # into the base class on getter access if the base class yields model
        # instances. That way, the queryable properties extensions stays
        # compatible to custom iterable classes while querysets can still be
        # pickled due to the base class being in the instance dict.
        cls = self.__dict__['_iterable_class']
        return QueryablePropertiesIterable.mix_with_class(cls, 'QueryableProperties' + cls.__name__)

    @_iterable_class.setter
    def _iterable_class(self, value):
        self.__dict__['_iterable_class'] = value

    def _clone(self, klass=None, *args, **kwargs):
        if klass:  # pragma: no cover
            # In older Django versions, the class of the queryset may be
            # replaced with a dynamically created class based on the current
            # class and the value of klass while cloning (e.g when using
            # .values()). Therefore this needs to be re-injected to be on top
            # of the MRO again to enable queryable properties functionality.
            klass = QueryablePropertiesQuerySetMixin.mix_with_class(klass, 'QueryableProperties' + klass.__name__)
            args = (klass,) + args
        clone = super(QueryablePropertiesQuerySetMixin, self)._clone(*args, **kwargs)
        # Since the _iterable_class property may return a dynamically created
        # class, the value of a clone must be reset to the base class.
        if '_iterable_class' in self.__dict__:
            clone._iterable_class = self.__dict__['_iterable_class']
        return clone

    def _resolve_update_kwargs(self, **kwargs):
        """
        Look for the names of queryable properties in the given keyword
        arguments for an update query and correctly resolve them into their
        actual keyword arguments.

        :param kwargs: Keyword arguments of an update query.
        :return: A dictionary containing the resolved arguments.
        :rtype: dict
        """
        original_names = set(kwargs)
        for original_name in original_names:
            try:
                prop = get_queryable_property(self.model, original_name)
            except QueryablePropertyDoesNotExist:
                continue
            if not prop.get_update_kwargs:
                raise QueryablePropertyError('Queryable property "{}" does not implement queryset updating.'
                                             .format(prop))

            # Call the method recursively since queryable properties can build
            # upon each other.
            additional_kwargs = self._resolve_update_kwargs(
                **prop.get_update_kwargs(self.model, kwargs.pop(original_name)))
            # Make sure that there are no conflicting values after resolving
            # the update keyword arguments of the queryable properties.
            for additional_name, value in six.iteritems(additional_kwargs):
                if additional_name in kwargs and kwargs[additional_name] != value:
                    raise QueryablePropertyError(
                        'Updating queryable property "{prop}" would change field "{field}", but a conflicting value '
                        'was set for this field by another queryable property or explicitly in the update arguments.'
                        .format(prop=prop, field=additional_name)
                    )
                kwargs[additional_name] = value

        return kwargs

    def select_properties(self, *names):
        """
        Add the annotations of the queryable properties with the specified
        names to this query. The annotation values will be cached in the
        properties of resulting model instances, regardless of the regular
        caching behavior of the queried properties.

        :param names: Names of queryable properties.
        :return: A copy of this queryset with the added annotations.
        :rtype: QuerySet
        """
        queryset = chain_queryset(self)
        for name in names:
            property_ref = QueryablePropertyReference(get_queryable_property(self.model, name), self.model, ())
            # A full GROUP BY is required if the query is not limited to
            # certain fields. Since only certain types of queries had the
            # _fields attribute in old Django versions, fall back to checking
            # for existing selection, on which the GROUP BY would be based.
            full_group_by = not getattr(self, '_fields', self.query.select)
            with queryset.query._add_queryable_property_annotation(property_ref, full_group_by, select=True):
                pass
        return queryset

    def iterator(self, *args, **kwargs):
        # Recent Django versions use the associated iterable class for the
        # iterator() implementation, where the QueryablePropertiesModelIterable
        # will be already mixed in. In older Django versions, use a standalone
        # QueryablePropertiesModelIterable instead to perform the queryable
        # properties processing (as long as this queryset returns model
        # instances).
        iterable = super(QueryablePropertiesQuerySetMixin, self).iterator(*args, **kwargs)
        if '_iterable_class' not in self.__dict__:  # pragma: no cover
            return iter(QueryablePropertiesIterable(self, iterable=iterable))
        return iterable

    def update(self, **kwargs):
        # Resolve any queryable properties into their actual update kwargs
        # before calling the base update method.
        kwargs = self._resolve_update_kwargs(**kwargs)
        return super(QueryablePropertiesQuerySetMixin, self).update(**kwargs)


class QueryablePropertiesQuerySet(QueryablePropertiesQuerySetMixin, QuerySet):
    """
    A special queryset class that allows to use queryable properties in its
    filter conditions, annotations and update queries.
    """
    pass


if hasattr(Manager, 'from_queryset'):
    QueryablePropertiesManager = Manager.from_queryset(QueryablePropertiesQuerySet)
else:  # pragma: no cover
    class QueryablePropertiesManager(Manager):

        def get_queryset(self):
            return QueryablePropertiesQuerySet(self.model, using=self._db)

        get_query_set = get_queryset

        def select_properties(self, *names):
            return self.get_queryset().select_properties(*names)



# encoding: utf-8
import pytest

from queryable_properties.query import QueryablePropertiesQueryMixin, QueryablePropertyReference
from tests.models import (ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties,
                          CategoryWithClassBasedProperties, CategoryWithDecoratorBasedProperties,
                          VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties)


class TestResolveQueryableProperty(object):

    @pytest.mark.parametrize('model, path, expected_property, expected_lookups', [
        # No relation involved
        (VersionWithClassBasedProperties, ['version'], VersionWithClassBasedProperties.version, []),
        (VersionWithDecoratorBasedProperties, ['version'], VersionWithDecoratorBasedProperties.version, []),
        (VersionWithClassBasedProperties, ['version', 'lower', 'exact'],
         VersionWithClassBasedProperties.version, ['lower', 'exact']),
        (VersionWithDecoratorBasedProperties, ['version', 'lower', 'exact'],
         VersionWithDecoratorBasedProperties.version, ['lower', 'exact']),
        # FK forward relation
        (VersionWithClassBasedProperties, ['application', 'version_count'],
         ApplicationWithClassBasedProperties.version_count, []),
        (VersionWithDecoratorBasedProperties, ['application', 'version_count'],
         ApplicationWithDecoratorBasedProperties.version_count, []),
        (VersionWithClassBasedProperties, ['application', 'major_sum', 'gt'],
         ApplicationWithClassBasedProperties.major_sum, ['gt']),
        (VersionWithDecoratorBasedProperties, ['application', 'major_sum', 'gt'],
         ApplicationWithDecoratorBasedProperties.major_sum, ['gt']),
        # FK reverse relation
        (ApplicationWithClassBasedProperties, ['versions', 'major_minor'],
         VersionWithClassBasedProperties.major_minor, []),
        (ApplicationWithDecoratorBasedProperties, ['versions', 'major_minor'],
         VersionWithDecoratorBasedProperties.major_minor, []),
        (ApplicationWithClassBasedProperties, ['versions', 'version', 'lower', 'contains'],
         VersionWithClassBasedProperties.version, ['lower', 'contains']),
        (ApplicationWithDecoratorBasedProperties, ['versions', 'version', 'lower', 'contains'],
         VersionWithDecoratorBasedProperties.version, ['lower', 'contains']),
        # M2M forward relation
        (ApplicationWithClassBasedProperties, ['categories', 'circular'],
         CategoryWithClassBasedProperties.circular, []),
        (ApplicationWithDecoratorBasedProperties, ['categories', 'circular'],
         CategoryWithDecoratorBasedProperties.circular, []),
        (ApplicationWithClassBasedProperties, ['categories', 'circular', 'exact'],
         CategoryWithClassBasedProperties.circular, ['exact']),
        (ApplicationWithDecoratorBasedProperties, ['categories', 'circular', 'exact'],
         CategoryWithDecoratorBasedProperties.circular, ['exact']),
        # M2M reverse relation
        (CategoryWithClassBasedProperties, ['applications', 'major_sum'],
         ApplicationWithClassBasedProperties.major_sum, []),
        (CategoryWithDecoratorBasedProperties, ['applications', 'major_sum'],
         ApplicationWithDecoratorBasedProperties.major_sum, []),
        (CategoryWithClassBasedProperties, ['applications', 'version_count', 'lt'],
         ApplicationWithClassBasedProperties.version_count, ['lt']),
        (CategoryWithDecoratorBasedProperties, ['applications', 'version_count', 'lt'],
         ApplicationWithDecoratorBasedProperties.version_count, ['lt']),
        # Multiple relations
        (CategoryWithClassBasedProperties, ['applications', 'versions', 'application', 'categories', 'circular'],
         CategoryWithClassBasedProperties.circular, []),
        (CategoryWithDecoratorBasedProperties, ['applications', 'versions', 'application', 'categories', 'circular'],
         CategoryWithDecoratorBasedProperties.circular, []),
        (VersionWithClassBasedProperties, ['application', 'categories', 'circular', 'in'],
         CategoryWithClassBasedProperties.circular, ['in']),
        (VersionWithDecoratorBasedProperties, ['application', 'categories', 'circular', 'in'],
         CategoryWithDecoratorBasedProperties.circular, ['in']),
    ])
    def test_successful(self, model, path, expected_property, expected_lookups):
        mixin = QueryablePropertiesQueryMixin()
        mixin.model = model
        expected_ref = QueryablePropertyReference(expected_property, expected_property.model,
                                                  tuple(path[:-len(expected_lookups) - 1]))
        assert mixin._resolve_queryable_property(path) == (expected_ref, expected_lookups)

    @pytest.mark.parametrize('model, path', [
        # No relation involved
        (VersionWithClassBasedProperties, ['non_existent']),
        (VersionWithDecoratorBasedProperties, ['non_existent']),
        (VersionWithClassBasedProperties, ['major']),
        (VersionWithDecoratorBasedProperties, ['major']),
        # FK forward relation
        (VersionWithClassBasedProperties, ['application', 'non_existent', 'exact']),
        (VersionWithDecoratorBasedProperties, ['application', 'non_existent', 'exact']),
        (VersionWithClassBasedProperties, ['application', 'name']),
        (VersionWithDecoratorBasedProperties, ['application', 'name']),
        # FK reverse relation
        (ApplicationWithClassBasedProperties, ['versions', 'non_existent']),
        (ApplicationWithDecoratorBasedProperties, ['versions', 'non_existent']),
        (ApplicationWithClassBasedProperties, ['versions', 'minor', 'gt']),
        (ApplicationWithDecoratorBasedProperties, ['versions', 'minor', 'gt']),
        # M2M forward relation
        (ApplicationWithClassBasedProperties, ['categories', 'non_existent']),
        (ApplicationWithDecoratorBasedProperties, ['categories', 'non_existent']),
        (ApplicationWithClassBasedProperties, ['categories', 'name']),
        (ApplicationWithDecoratorBasedProperties, ['categories', 'name']),
        # M2M reverse relation
        (CategoryWithClassBasedProperties, ['applications', 'non_existent']),
        (CategoryWithDecoratorBasedProperties, ['applications', 'non_existent']),
        (CategoryWithClassBasedProperties, ['applications', 'name']),
        (CategoryWithDecoratorBasedProperties, ['applications', 'name']),
        # Non existent relation
        (VersionWithClassBasedProperties, ['non_existent_relation', 'non_existent', 'in']),
        (VersionWithDecoratorBasedProperties, ['non_existent_relation', 'non_existent', 'in']),
    ])
    def test_unsuccessful(self, model, path):
        mixin = QueryablePropertiesQueryMixin()
        mixin.model = model
        assert mixin._resolve_queryable_property(path) == (None, [])



# encoding: utf-8
"""Tests for basic features of the Query and QuerySet mixins."""

import pytest

from six.moves import cPickle

from .models import ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties


@pytest.mark.django_db
@pytest.mark.usefixtures('versions')
@pytest.mark.parametrize('model', [ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties])
def test_pickle_unpickle(model):
    queryset1 = model.objects.filter(version_count=4).order_by('name').select_properties('version_count')
    expected_applications = list(queryset1)
    serialized_query = cPickle.dumps(queryset1.query)
    queryset2 = model.objects.all()
    queryset2.query = cPickle.loads(serialized_query)
    serialized_queryset = cPickle.dumps(queryset1)
    queryset3 = cPickle.loads(serialized_queryset)
    for queryset in (queryset1, queryset2, queryset3):
        versions = list(queryset)
        assert versions == expected_applications
        assert all(model.version_count._has_cached_value(obj) for obj in queryset)



# encoding: utf-8

import pytest

from django import VERSION as DJANGO_VERSION
from django.db import models

from queryable_properties.exceptions import FieldError, QueryablePropertyError

from ..models import (ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties,
                      CategoryWithClassBasedProperties, CategoryWithDecoratorBasedProperties,
                      VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties)

pytestmark = pytest.mark.django_db


class TestFilter(object):

    @pytest.mark.parametrize('model', [VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties])
    def test_exception_on_unimplemented_filter(self, monkeypatch, model):
        monkeypatch.setattr(model.version, 'get_filter', None)
        with pytest.raises(QueryablePropertyError):
            model.objects.filter(version='1.2.3')

    @pytest.mark.parametrize('model', [VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties])
    def test_standard_exception_on_invalid_field_name(self, model):
        with pytest.raises(FieldError):
            model.objects.filter(non_existent_field=1337)

    @pytest.mark.skipif(DJANGO_VERSION < (1, 9), reason="type check didn't exist before Django 1.9")
    @pytest.mark.parametrize('model', [VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties])
    def test_standard_exception_on_invalid_filter_expression(self, model):
        with pytest.raises(FieldError):
            # The dict is passed as arg instead of kwargs, making it an invalid
            # filter expression.
            model.objects.filter(models.Q({'version': '2.0.0'}))


class TestAnnotation(object):

    @pytest.mark.parametrize('model', [VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties])
    def test_exception_on_unimplemented_annotater(self, model):
        with pytest.raises(QueryablePropertyError):
            model.objects.select_properties('major_minor')

    @pytest.mark.skipif(DJANGO_VERSION < (1, 8), reason="Expression-based annotations didn't exist before Django 1.8")
    @pytest.mark.parametrize('model', [CategoryWithClassBasedProperties, CategoryWithDecoratorBasedProperties])
    def test_circular_property(self, model):
        with pytest.raises(QueryablePropertyError, match='circular dependency'):
            model.objects.filter(circular=1337)
        with pytest.raises(QueryablePropertyError, match='circular dependency'):
            model.objects.select_properties('circular')


class TestUpdate(object):

    @pytest.mark.parametrize('model', [ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties])
    def test_exception_on_unimplemented_updater(self, model):
        with pytest.raises(QueryablePropertyError):
            model.objects.update(highest_version='1.3.37')

    @pytest.mark.parametrize('model, kwargs', [
        (VersionWithClassBasedProperties, {'major_minor': '42.42', 'major': 18}),
        (VersionWithClassBasedProperties, {'major_minor': '1.2', 'version': '1.3.37', 'minor': 5}),
        (VersionWithDecoratorBasedProperties, {'major_minor': '42.42', 'major': 18}),
        (VersionWithDecoratorBasedProperties, {'major_minor': '1.2', 'version': '1.3.37', 'minor': 5}),
    ])
    def test_exception_on_conflicting_values(self, model, kwargs):
        with pytest.raises(QueryablePropertyError):
            model.objects.update(**kwargs)


class TestOrdering(object):

    @pytest.mark.parametrize('model', [VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties])
    def test_exception_on_unimplemented_annotater(self, model):
        with pytest.raises(QueryablePropertyError):
            iter(model.objects.order_by('major_minor'))



# encoding: utf-8

from datetime import date

import pytest
from mock import Mock

try:
    from django.db.models import Value
except ImportError:
    Value = Mock()
try:
    from django.db.models.functions import Concat
except ImportError:
    Concat = Mock()

from .models import (ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties,
                     CategoryWithClassBasedProperties, CategoryWithDecoratorBasedProperties, Version)


@pytest.fixture
def categories():
    return [
        CategoryWithClassBasedProperties.objects.create(name='Linux apps'),
        CategoryWithClassBasedProperties.objects.create(name='Windows apps'),
        CategoryWithDecoratorBasedProperties.objects.create(name='Linux apps'),
        CategoryWithDecoratorBasedProperties.objects.create(name='Windows apps'),
    ]


@pytest.fixture
def applications(categories):
    apps = [
        ApplicationWithClassBasedProperties.objects.create(name='My cool App'),
        ApplicationWithClassBasedProperties.objects.create(name='Another App'),
        ApplicationWithDecoratorBasedProperties.objects.create(name='My cool App'),
        ApplicationWithDecoratorBasedProperties.objects.create(name='Another App'),
    ]
    apps[0].categories.add(categories[0])
    apps[1].categories.add(categories[0])
    apps[1].categories.add(categories[1])
    apps[2].categories.add(categories[2])
    apps[3].categories.add(categories[2])
    apps[3].categories.add(categories[3])
    return apps


@pytest.fixture
def versions(applications):
    objs = []
    for application in applications:
        objs.extend([
            application.versions.create(major=1, minor=2, patch=3, release_type=Version.BETA,
                                        supported_until=date(2016, 12, 31)),
            application.versions.create(major=1, minor=3, patch=0,
                                        supported_from=date(2017, 1, 1), supported_until=date(2017, 12, 31)),
            application.versions.create(major=1, minor=3, patch=1,
                                        supported_from=date(2018, 1, 1), supported_until=date(2018, 12, 31)),
            application.versions.create(major=2, minor=0, patch=0, changes='Amazing new features',
                                        release_type=Version.ALPHA, supported_from=date(2018, 11, 1)),
        ])
    return objs



# encoding: utf-8
import pytest

from django import VERSION as DJANGO_VERSION
from django.db.models import Q

from queryable_properties.exceptions import QueryablePropertyError
from queryable_properties.properties import (AnnotationMixin, boolean_filter, LookupFilterMixin, lookup_filter,
                                             QueryableProperty)

from ..models import ApplicationWithClassBasedProperties


class BaseLookupFilterProperty(LookupFilterMixin, QueryableProperty):

    @boolean_filter
    def filter_boolean(self, cls):
        return Q(dummy__year__gt=2000)

    @lookup_filter('iexact')
    def filter_ci_equality(self, cls, lookup, value):
        return Q(dummy=value)

    @lookup_filter('lt', 'lte')
    def filter_lower(self, cls, lookup, value):
        return Q(**{'dummy__day__{}'.format(lookup): value})


class DerivedLookupFilterProperty(BaseLookupFilterProperty):

    @lookup_filter('lt', 'gt')
    def filter_comparison(self, cls, lookup, value):
        return Q(**{'dummy__month__{}'.format(lookup): value})

    @lookup_filter('in')
    def filter_in(self, cls, lookup, value):
        value = list(value) + ['test']
        return Q(dummy__in=value)


class TestLookupFilterMixin(object):

    def test_registration(self):
        base = BaseLookupFilterProperty()
        assert set(base.lookup_mappings) == {'exact', 'iexact', 'lt', 'lte'}
        assert base.lookup_mappings['exact'] != base.lookup_mappings['iexact']
        assert base.lookup_mappings['exact'] != base.lookup_mappings['lt']
        assert base.lookup_mappings['iexact'] != base.lookup_mappings['lt']
        assert base.lookup_mappings['lte'] == base.lookup_mappings['lt']

        derived = DerivedLookupFilterProperty()
        assert set(derived.lookup_mappings) == {'exact', 'iexact', 'lt', 'lte', 'gt', 'in'}
        assert derived.lookup_mappings['exact'] != derived.lookup_mappings['lt']
        assert derived.lookup_mappings['exact'] != derived.lookup_mappings['iexact']
        assert derived.lookup_mappings['iexact'] != derived.lookup_mappings['lt']
        assert derived.lookup_mappings['lte'] != derived.lookup_mappings['lt']
        assert derived.lookup_mappings['gt'] == derived.lookup_mappings['lt']

    @pytest.mark.parametrize('cls, lookup, value, expected_q_value, expected_q_negation', [
        (BaseLookupFilterProperty, 'exact', True, ('dummy__year__gt', 2000), False),
        (BaseLookupFilterProperty, 'exact', False, ('dummy__year__gt', 2000), True),
        (BaseLookupFilterProperty, 'iexact', 'test', ('dummy', 'test'), False),
        (BaseLookupFilterProperty, 'lt', 42, ('dummy__day__lt', 42), False),
        (BaseLookupFilterProperty, 'lte', 1337, ('dummy__day__lte', 1337), False),

        (DerivedLookupFilterProperty, 'exact', True, ('dummy__year__gt', 2000), False),
        (DerivedLookupFilterProperty, 'exact', False, ('dummy__year__gt', 2000), True),
        (DerivedLookupFilterProperty, 'iexact', 'test', ('dummy', 'test'), False),
        (DerivedLookupFilterProperty, 'lt', 42, ('dummy__month__lt', 42), False),
        (DerivedLookupFilterProperty, 'gt', 69, ('dummy__month__gt', 69), False),
        (DerivedLookupFilterProperty, 'lte', 1337, ('dummy__day__lte', 1337), False),
        (DerivedLookupFilterProperty, 'in', ('a', 'b'), ('dummy__in', ['a', 'b', 'test']), False),
    ])
    def test_filter_call(self, cls, lookup, value, expected_q_value, expected_q_negation):
        prop = cls()
        q = prop.get_filter(None, lookup, value)
        if DJANGO_VERSION < (1, 6) and expected_q_negation:
            # In very old Django versions, negating adds another layer.
            q = q.children[0]
        assert len(q.children) == 1
        assert q.children[0] == expected_q_value
        assert q.negated is expected_q_negation

    @pytest.mark.parametrize('cls, lookup, value', [
        (BaseLookupFilterProperty, 'month', 5),
        (BaseLookupFilterProperty, 'gt', 42),
        (BaseLookupFilterProperty, 'gte', 42),
        (BaseLookupFilterProperty, 'in', ['a', 'b']),
        (DerivedLookupFilterProperty, 'month', 5),
        (DerivedLookupFilterProperty, 'gte', 42),
    ])
    def test_filter_call_not_implemented(self, cls, lookup, value):
        prop = cls()
        prop.model = ApplicationWithClassBasedProperties
        prop.name = 'dummy'
        with pytest.raises(QueryablePropertyError):
            prop.get_filter(None, lookup, value)


class TestAnnotationMixin(object):

    @pytest.fixture
    def prop(self):
        cls = AnnotationMixin.mix_with_class(QueryableProperty)
        prop = cls()
        prop.name = 'test'
        return prop

    @pytest.mark.parametrize('lookup, value', [
        ('exact', 'abc'),
        ('isnull', True),
        ('lte', 5),
    ])
    def test_get_filter(self, prop, lookup, value):
        q = prop.get_filter(ApplicationWithClassBasedProperties, lookup, value)
        assert isinstance(q, Q)
        assert len(q.children) == 1
        assert q.children[0] == ('test__{}'.format(lookup), value)



# encoding: utf-8

from functools import wraps

import six
from django.db.models import Q

from ..compat import LOOKUP_SEP
from ..exceptions import QueryablePropertyError
from ..utils import InjectableMixin


class LookupFilterMeta(type):
    """
    Metaclass for classes that use the :class:`LookupFilterMixin` to detect the
    individual registered filter methods and make them available to the main
    filter method.
    """

    def __new__(mcs, name, bases, attrs):
        # Find all methods that have been marked with lookups via the
        # `lookup_filter` decorator.
        mappings = {}
        for attr_name, attr in six.iteritems(attrs):
            if callable(attr) and hasattr(attr, '_lookups'):
                for lookup in attr._lookups:
                    mappings[lookup] = attr_name

        # Let the class construction take care of the lookup mappings of the
        # base class(es) and add the ones from the current class to them.
        cls = super(LookupFilterMeta, mcs).__new__(mcs, name, bases, attrs)
        cls._lookup_mappings = dict(cls._lookup_mappings, **mappings)
        return cls


class LookupFilterMixin(six.with_metaclass(LookupFilterMeta, InjectableMixin)):
    """
    A mixin for queryable properties that allows to implement queryset
    filtering via individual methods for different lookups.
    """

    # Avoid overriding the __reduce__ implementation of queryable properties.
    _dynamic_pickling = False

    # Stores mappings of lookups to the names of their corresponding filter
    # functions.
    _lookup_mappings = {}

    def __init__(self, *args, **kwargs):
        self.lookup_mappings = {lookup: getattr(self, name) for lookup, name in six.iteritems(self._lookup_mappings)}
        super(LookupFilterMixin, self).__init__(*args, **kwargs)

    @classmethod
    def lookup_filter(cls, *lookups):
        """
        Decorator for individual filter methods of classes that use the
        :class:`LookupFilterMixin` to register the decorated methods for the
        given lookups.

        :param str lookups: The lookups to register the decorated method for.
        :return: The actual internal decorator.
        :rtype: function
        """
        def decorator(func):
            func._lookups = lookups  # Store the lookups on the function to be able to read them in the meta class.
            return func
        return decorator

    @classmethod
    def boolean_filter(cls, method):
        """
        Decorator for individual filter methods of classes that use the
        :class:`LookupFilterMixin` to register the methods that are simple
        boolean filters (i.e. the filter can only be called with a `True` or
        `False` value). This automatically restricts the usable lookups to
        `exact`. Decorated methods should not expect the `lookup` and `value`
        parameters and should always return a `Q` object representing the
        positive (i.e. `True`) filter case. The decorator will automatically
        negate the condition if the filter was called with a `False` value.

        :param function method: The method to decorate.
        :return: The decorated method.
        :rtype: function
        """
        @wraps(method)
        def filter_wrapper(self, model, lookup, value):
            """Actual filter method that negates the condition if required."""
            condition = method(self, model)
            if not value:
                condition.negate()
            return condition
        lookup_decorator = cls.lookup_filter('exact')
        return lookup_decorator(filter_wrapper)

    def get_filter(self, cls, lookup, value):
        method = self.lookup_mappings.get(lookup)
        if not method:
            raise QueryablePropertyError(
                'Queryable property "{prop}" does not implement filtering with lookup "{lookup}".'
                .format(prop=self, lookup=lookup)
            )
        return method(cls, lookup, value)


# Aliases to allow the usage of the decorators without the "LookupFilterMixin."
# prefix.
boolean_filter = LookupFilterMixin.boolean_filter
lookup_filter = LookupFilterMixin.lookup_filter


class SetterMixin(object):
    """
    A mixin for queryable properties that also define a setter.
    """

    def set_value(self, obj, value):  # pragma: no cover
        """
        Setter method for the queryable property, which will be called when the
        property is write-accessed.

        :param django.db.models.Model obj: The object on which the property was
                                           accessed.
        :param value: The value to set.
        """
        raise NotImplementedError()


class AnnotationMixin(InjectableMixin):
    """
    A mixin for queryable properties that allows to add an annotation to
    represent them to querysets.
    """

    # Avoid overriding the __reduce__ implementation of queryable properties.
    _dynamic_pickling = False

    filter_requires_annotation = True

    def get_annotation(self, cls):  # pragma: no cover
        """
        Construct an annotation representing this property that can be added
        to querysets of the model associated with this property.

        :param type cls: The model class of which a queryset should be
                         annotated.
        :return: An annotation object.
        """
        raise NotImplementedError()

    def get_filter(self, cls, lookup, value):
        # Since annotations can be filtered like regular fields, a Q object
        # that simply passes the filter through can be used.
        return Q(**{LOOKUP_SEP.join((self.name, lookup)): value})


class UpdateMixin(object):
    """
    A mixin for queryable properties that allows to use themselves in update
    queries.
    """

    def get_update_kwargs(self, cls, value):  # pragma: no cover
        """
        Resolve an update keyword argument for this property into the actual
        keyword arguments to emulate an update using this property.

        :param type cls: The model class of which an update query should be
                         performed.
        :param value: The value passed to the update call for this property.
        :return: The actual keyword arguments to set in the update call instead
                 of the given one.
        :rtype: dict
        """
        raise NotImplementedError()



# encoding: utf-8
import pytest

from django.db.models import Q
from six.moves import cPickle

from queryable_properties.exceptions import QueryablePropertyDoesNotExist
from queryable_properties.properties import QueryableProperty
from queryable_properties.utils import (get_queryable_property, InjectableMixin, MISSING_OBJECT, ModelAttributeGetter,
                                        parametrizable_decorator, TreeNodeProcessor)

from .models import VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties


class DummyClass(object):

    def __init__(self, attr1, attr2):
        self.attr1 = attr1
        self.attr2 = attr2

    @parametrizable_decorator
    def decorator(self, function, *args, **kwargs):
        function.args = args
        function.kwargs = kwargs
        return function


class DummyMixin(InjectableMixin):

    def init_injected_attrs(self):
        self.mixin_attr1 = 1.337
        self.mixin_attr2 = 'test'


class TestGetQueryableProperty(object):

    @pytest.mark.parametrize('model, property_name', [
        (VersionWithClassBasedProperties, 'major_minor'),
        (VersionWithDecoratorBasedProperties, 'major_minor'),
        (VersionWithClassBasedProperties, 'version'),
        (VersionWithDecoratorBasedProperties, 'version'),
    ])
    def test_property_found(self, model, property_name):
        prop = get_queryable_property(model, property_name)
        assert isinstance(prop, QueryableProperty)

    @pytest.mark.parametrize('model, property_name', [
        (VersionWithClassBasedProperties, 'non_existent'),
        (VersionWithDecoratorBasedProperties, 'non_existent'),
        (VersionWithClassBasedProperties, 'major'),  # Existing model field
        (VersionWithDecoratorBasedProperties, 'major'),  # Existing model field
    ])
    def test_exception(self, model, property_name):
        with pytest.raises(QueryablePropertyDoesNotExist):
            get_queryable_property(model, property_name)


class TestInjectableMixin(object):

    @pytest.mark.parametrize('class_name, expected_class_name', [
        (None, DummyClass.__name__),
        ('TestClass', 'TestClass'),
    ])
    def test_mix_with_class(self, monkeypatch, class_name, expected_class_name):
        monkeypatch.setattr(DummyMixin, '_created_classes', {})
        assert not DummyMixin._created_classes
        created_classes = set()

        # Execute the code twice to test the cache
        for _ in range(2):
            cls = DummyMixin.mix_with_class(DummyClass, class_name)
            created_classes.add(cls)
            assert issubclass(cls, DummyClass)
            assert issubclass(cls, DummyMixin)
            assert cls.__name__ == expected_class_name
            assert len(DummyMixin._created_classes) == 1
            assert len(created_classes) == 1
            # Test that the __init__ method of the new class correctly
            # initializes the injected attributes.
            obj = cls(5, 'abc')
            assert obj.mixin_attr1 == 1.337
            assert obj.mixin_attr2 == 'test'

    def test_inject_into_object(self):
        obj = DummyClass(5, 'abc')
        DummyMixin.inject_into_object(obj)
        assert isinstance(obj, DummyClass)
        assert isinstance(obj, DummyMixin)
        assert obj.attr1 == 5
        assert obj.attr2 == 'abc'
        assert obj.mixin_attr1 == 1.337
        assert obj.mixin_attr2 == 'test'

    def test_pickle_unpickle(self):
        base_obj = DummyClass('xyz', 42.42)
        DummyMixin.inject_into_object(base_obj)
        serialized_obj = cPickle.dumps(base_obj)
        deserialized_obj = cPickle.loads(serialized_obj)

        for obj in (base_obj, deserialized_obj):
            assert isinstance(obj, DummyClass)
            assert isinstance(obj, DummyMixin)
            assert obj.attr1 == 'xyz'
            assert obj.attr2 == 42.42
            assert obj.mixin_attr1 == 1.337
            assert obj.mixin_attr2 == 'test'

    def test_no_reduce_implementation(self, monkeypatch):
        monkeypatch.setattr(DummyMixin, '_dynamic_pickling', False)
        monkeypatch.setattr(DummyMixin, '_created_classes', {})
        base_obj = DummyClass('xyz', 42.42)
        DummyMixin.inject_into_object(base_obj)
        with pytest.raises(cPickle.PicklingError):
            cPickle.dumps(base_obj)


class TestTreeNodeProcessor(object):

    @pytest.mark.parametrize('node, expected_result', [
        (Q(a=1), True),
        (Q(b=2), False),
        (Q(Q(a=1) | Q(b=2), c=3), True),
        (Q(Q(d=1) | Q(b=2), c=3), False),
    ])
    def test_check_leaves(self, node, expected_result):
        # The predicate checks if a leaf for field 'a' exists
        assert TreeNodeProcessor(node).check_leaves(lambda item: item[0] == 'a') is expected_result

    @pytest.mark.parametrize('copy', [True, False])
    def test_modify_tree_node(self, copy):
        q = Q(Q(a=1) | Q(b=2), c=3)
        result = TreeNodeProcessor(q).modify_leaves(
            lambda item: ('prefix_{}_suffix'.format(item[0]), item[1] + 1), copy=copy)
        assert (result is q) is not copy
        children = list(result.children)
        assert ('prefix_c_suffix', 4) in children
        children.remove(('prefix_c_suffix', 4))
        assert ('prefix_a_suffix', 2) in children[0].children
        assert ('prefix_b_suffix', 3) in children[0].children


def test_parametrizable_decorator():
    dummy = DummyClass(1, 2)

    @dummy.decorator
    def func1():
        pass

    @dummy.decorator(some_kwarg=1, another_kwarg='test')
    def func2():
        pass

    def func3():
        pass

    func3 = dummy.decorator(func3, 1, 2, kwarg='test')

    assert func1.args == ()
    assert func1.kwargs == {}
    assert func2.args == ()
    assert func2.kwargs == dict(some_kwarg=1, another_kwarg='test')
    assert func3.args == (1, 2)
    assert func3.kwargs == dict(kwarg='test')


class TestModelAttributeGetter(object):

    @pytest.mark.parametrize('path, expected_parts', [
        ('attr', ['attr']),
        ('attr1.attr2', ['attr1', 'attr2']),
        ('attr1.attr2.attr3', ['attr1', 'attr2', 'attr3']),
    ])
    def test_initializer(self, path, expected_parts):
        getter = ModelAttributeGetter(path)
        assert getter.path_parts == expected_parts

    @pytest.mark.django_db
    @pytest.mark.parametrize('path, expected_value', [
        ('major', 1),
        ('changes', None),
        ('application.name', 'My cool App'),
    ])
    def test_get_value(self, versions, path, expected_value):
        obj = versions[0]
        getter = ModelAttributeGetter(path)
        assert getter.get_value(obj) == expected_value

    @pytest.mark.django_db
    def test_get_value_catch_attribute_error_on_none(self, versions):
        obj = versions[0]
        getter = ModelAttributeGetter('changes.strip')
        assert getter.get_value(obj) is MISSING_OBJECT

    @pytest.mark.django_db
    @pytest.mark.parametrize('path', ['non_existent', 'non.existent', 'application.non_existent'])
    def test_get_value_bubble_attribute_error(self, versions, path):
        obj = versions[0]
        getter = ModelAttributeGetter(path)
        with pytest.raises(AttributeError):
            getter.get_value(obj)

    @pytest.mark.django_db
    def test_get_value_catch_object_does_not_exist(self, applications, versions):
        obj = versions[0].__class__.objects.get(pk=versions[0].pk)  # Refresh from DB
        applications[0].delete()
        getter = ModelAttributeGetter('application.name')
        assert getter.get_value(obj) is MISSING_OBJECT

    @pytest.mark.parametrize('path, lookup, value, expected_query_name', [
        ('attr', 'exact', 1337, 'attr__exact'),
        ('attr1.attr2', 'in', ('test', 'value'), 'attr1__attr2__in'),
        ('attr1.attr2.attr3', 'gt', 1337, 'attr1__attr2__attr3__gt'),
    ])
    def test_build_filter(self, path, lookup, value, expected_query_name):
        getter = ModelAttributeGetter(path)
        condition = getter.build_filter(lookup, value)
        assert isinstance(condition, Q)
        assert condition.children[0] == (expected_query_name, value)



# encoding: utf-8

from datetime import date
from itertools import chain

import pytest

from django import VERSION as DJANGO_VERSION
from django.db.models import Q

from ..models import ApplicationWithClassBasedProperties, VersionWithClassBasedProperties

pytestmark = [pytest.mark.django_db, pytest.mark.usefixtures('versions')]


class TestValueCheckProperty(object):

    @pytest.mark.parametrize('index, is_alpha, is_beta, is_stable, is_unstable, released_in_2018', [
        (0, False, True, False, True, False),
        (1, False, False, True, False, False),
        (2, False, False, True, False, True),
        (3, True, False, False, True, True),
    ])
    def test_getter(self, versions, index, is_alpha, is_beta, is_stable, is_unstable, released_in_2018):
        version = versions[index]
        assert version.is_alpha is is_alpha
        assert version.is_beta is is_beta
        assert version.is_stable is is_stable
        assert version.is_unstable is is_unstable
        assert version.shares_common_data is True
        assert version.released_in_2018 is released_in_2018

    @pytest.mark.parametrize('condition, expected_versions', [
        (Q(is_alpha=True), {'2.0.0'}),
        (Q(is_beta=True), {'1.2.3'}),
        (Q(is_stable=True), {'1.3.0', '1.3.1'}),
        (Q(is_unstable=True), {'1.2.3', '2.0.0'}),
        (Q(is_alpha=True) | Q(is_beta=True), {'1.2.3', '2.0.0'}),
        (Q(is_stable=False), {'1.2.3', '2.0.0'}),
        (Q(is_alpha=False), {'1.2.3', '1.3.0', '1.3.1'}),
        (Q(shares_common_data=False), set()),
    ])
    def test_filter(self, condition, expected_versions):
        results = VersionWithClassBasedProperties.objects.filter(condition)
        assert len(results) == len(expected_versions) * 2
        assert set(result.version for result in results) == expected_versions

    @pytest.mark.skipif(DJANGO_VERSION < (1, 9), reason="Transforms and lookup couldn't be combined before Django 1.9")
    @pytest.mark.parametrize('condition, expected_versions', [
        (Q(released_in_2018=True), {'1.3.1', '2.0.0'}),
        (Q(released_in_2018=True, is_alpha=True), {'2.0.0'}),
        (Q(released_in_2018=False), {'1.2.3', '1.3.0'}),
    ])
    def test_filter_based_on_transform(self, condition, expected_versions):
        results = VersionWithClassBasedProperties.objects.filter(condition)
        assert len(results) == len(expected_versions) * 2
        assert set(result.version for result in results) == expected_versions

    @pytest.mark.skipif(DJANGO_VERSION < (1, 8), reason="Expression-based annotations didn't exist before Django 1.8")
    @pytest.mark.parametrize('ordering, expected_version_order', [
        (('-is_stable', '-version'), ['1.3.1', '1.3.0', '2.0.0', '1.2.3']),
        (('-is_unstable', '-is_alpha', 'version'), ['2.0.0', '1.2.3', '1.3.0', '1.3.1']),
    ])
    def test_annotation(self, ordering, expected_version_order):
        results = VersionWithClassBasedProperties.objects.order_by(*ordering)
        # There are 2 objects for each version number.
        expected_version_order = list(chain(*zip(expected_version_order, expected_version_order)))
        assert [result.version for result in results] == expected_version_order

    @pytest.mark.skipif(DJANGO_VERSION < (1, 9), reason="Transforms and lookup couldn't be combined before Django 1.9")
    def test_annotation_based_on_transform(self):
        results = VersionWithClassBasedProperties.objects.order_by('released_in_2018', '-version')
        assert [result.version for result in results] == [
            '1.3.0', '1.3.0', '1.2.3', '1.2.3', '2.0.0', '2.0.0', '1.3.1', '1.3.1'
        ]


class TestRangeCheckProperty(object):

    def test_final_value(self, monkeypatch):
        assert VersionWithClassBasedProperties.is_supported.final_value == date(2019, 1, 1)
        monkeypatch.setattr(VersionWithClassBasedProperties.is_supported, 'value', lambda: 5)
        assert VersionWithClassBasedProperties.is_supported.final_value == 5

    @pytest.mark.parametrize(
        'index, prop_name, value, include_boundaries, include_missing, in_range, expected_result',
        [
            (3, 'is_supported', date(2019, 1, 1), True, True, True, True),
            (3, 'is_supported', date(2019, 1, 1), True, True, False, False),
            (3, 'is_supported', date(2019, 1, 1), True, False, True, False),
            (0, 'supported_in_2018', 2018, True, True, True, False),
            (3, 'supported_in_2018', 2018, False, True, True, False),
            (0, 'supported_in_2018', 2018, True, True, False, True),
            (3, 'supported_in_2018', 2018, True, False, False, True),
            (0, 'is_supported', date(2016, 12, 31), True, True, True, True),
            (0, 'is_supported', date(2016, 12, 31), True, True, False, False),
            (0, 'is_supported', date(2016, 12, 31), False, True, True, False),
            (0, 'is_supported', date(2016, 12, 31), False, True, False, True),
            (3, 'is_supported', date(2018, 11, 1), True, True, True, True),
            (3, 'is_supported', date(2018, 11, 1), True, True, False, False),
            (3, 'is_supported', date(2018, 11, 1), False, True, True, False),
            (3, 'is_supported', date(2018, 11, 1), False, True, False, True),
        ]
    )
    def test_getter(self, monkeypatch, versions, index, prop_name, value, include_boundaries, include_missing,
                    in_range, expected_result):
        version = versions[index]
        prop = getattr(VersionWithClassBasedProperties, prop_name)
        monkeypatch.setattr(prop, 'value', value)
        monkeypatch.setattr(prop, 'include_boundaries', include_boundaries)
        monkeypatch.setattr(prop, 'include_missing', include_missing)
        monkeypatch.setattr(prop, 'in_range', in_range)
        assert getattr(version, prop_name) is expected_result

    @pytest.mark.parametrize('value, include_boundaries, include_missing, in_range, condition, expected_versions', [
        (date(2019, 1, 1), True, True, True, Q(is_supported=True), {'2.0.0'}),
        (date(2019, 1, 1), True, True, True, Q(is_supported=True, major=1), set()),
        (date(2019, 1, 1), False, True, True, Q(is_supported=True), {'2.0.0'}),
        (date(2019, 1, 1), True, True, True, Q(is_supported=False), {'1.2.3', '1.3.0', '1.3.1'}),
        (date(2019, 1, 1), True, True, False, Q(is_supported=True), {'1.2.3', '1.3.0', '1.3.1'}),
        (date(2019, 1, 1), True, True, False, Q(is_supported=False), {'2.0.0'}),
        (date(2019, 1, 1), True, False, True, Q(is_supported=True), set()),
        (date(2019, 1, 1), True, False, False, Q(is_supported=True), {'1.2.3', '1.3.0', '1.3.1', '2.0.0'}),
        (date(2016, 12, 31), True, True, True, Q(is_supported=True), {'1.2.3'}),
        (date(2016, 12, 31), True, True, False, Q(is_supported=True), {'1.3.0', '1.3.1', '2.0.0'}),
        (date(2016, 12, 31), False, True, True, Q(is_supported=True), set()),
        (date(2016, 12, 31), False, True, False, Q(is_supported=True), {'1.2.3', '1.3.0', '1.3.1', '2.0.0'}),
        (date(2018, 11, 1), True, True, True, Q(is_supported=True), {'1.3.1', '2.0.0'}),
        (date(2018, 11, 1), True, True, False, Q(is_supported=True), {'1.2.3', '1.3.0'}),
        (date(2018, 11, 1), False, True, True, Q(is_supported=True), {'1.3.1'}),
        (date(2018, 11, 1), False, True, False, Q(is_supported=True), {'1.2.3', '1.3.0', '2.0.0'}),
    ])
    def test_filter(self, monkeypatch, value, include_boundaries, include_missing, in_range, condition,
                    expected_versions):
        prop = VersionWithClassBasedProperties.is_supported
        monkeypatch.setattr(prop, 'value', value)
        monkeypatch.setattr(prop, 'include_boundaries', include_boundaries)
        monkeypatch.setattr(prop, 'include_missing', include_missing)
        monkeypatch.setattr(prop, 'in_range', in_range)
        results = VersionWithClassBasedProperties.objects.filter(condition)
        assert set(version.version for version in results) == expected_versions

    @pytest.mark.skipif(DJANGO_VERSION < (1, 9), reason="Transforms and lookup couldn't be combined before Django 1.9")
    @pytest.mark.parametrize('include_boundaries, include_missing, in_range, condition, expected_versions', [
        (True, True, True, Q(supported_in_2018=True), {'1.3.1', '2.0.0'}),
        (True, True, True, Q(supported_in_2018=True, major=1), {'1.3.1'}),
        (True, True, True, Q(supported_in_2018=False), {'1.2.3', '1.3.0'}),
        (True, False, True, Q(supported_in_2018=True), {'1.3.1'}),
        (False, True, True, Q(supported_in_2018=True), set()),
        (True, True, False, Q(supported_in_2018=True), {'1.2.3', '1.3.0'}),
        (True, True, False, Q(supported_in_2018=False), {'1.3.1', '2.0.0'}),
        (True, False, False, Q(supported_in_2018=True), {'1.2.3', '1.3.0', '2.0.0'}),
    ])
    def test_filter_based_on_transform(self, monkeypatch, include_boundaries, include_missing, in_range, condition,
                                       expected_versions):
        prop = VersionWithClassBasedProperties.supported_in_2018
        monkeypatch.setattr(prop, 'include_boundaries', include_boundaries)
        monkeypatch.setattr(prop, 'include_missing', include_missing)
        monkeypatch.setattr(prop, 'in_range', in_range)
        results = VersionWithClassBasedProperties.objects.filter(condition)
        assert set(version.version for version in results) == expected_versions

    @pytest.mark.skipif(DJANGO_VERSION < (1, 8), reason="Expression-based annotations didn't exist before Django 1.8")
    def test_annotation(self):
        results = VersionWithClassBasedProperties.objects.order_by('-is_supported', 'version')
        assert list(results.select_properties('version').values_list('version', flat=True)) == [
            '2.0.0', '2.0.0', '1.2.3', '1.2.3', '1.3.0', '1.3.0', '1.3.1', '1.3.1'
        ]

    @pytest.mark.skipif(DJANGO_VERSION < (1, 9), reason="Transforms and lookup couldn't be combined before Django 1.9")
    def test_annotation_based_on_transform(self):
        results = VersionWithClassBasedProperties.objects.order_by('-supported_in_2018', 'version')
        assert list(results.select_properties('version').values_list('version', flat=True)) == [
            '1.3.1', '1.3.1', '2.0.0', '2.0.0', '1.2.3', '1.2.3', '1.3.0', '1.3.0'
        ]


class TestAggregateProperty(object):

    def test_getter(self, applications, versions):
        assert applications[0].major_sum == 5
        versions[0].delete()
        assert applications[0].major_sum == 4

    def test_annotation(self, applications, versions):
        versions[0].delete()
        assert list(ApplicationWithClassBasedProperties.objects.order_by('-major_sum')) == [
            applications[1], applications[0]
        ]



# encoding: utf-8
import pytest

import six
from django import VERSION as DJANGO_VERSION
from django.db.models import F, Model, Q

from queryable_properties.exceptions import QueryablePropertyError
from queryable_properties.properties import (AnnotationMixin, CACHE_RETURN_VALUE, CACHE_VALUE, CLEAR_CACHE, DO_NOTHING,
                                             LookupFilterMixin, QueryableProperty, queryable_property)
from queryable_properties.utils import reset_queryable_property

from ..models import (ApplicationWithClassBasedProperties, Category, DummyProperty, VersionWithClassBasedProperties,
                      VersionWithDecoratorBasedProperties)


def function_with_docstring():
    """Just a dummy function."""
    pass


class TestBasics(object):

    @pytest.fixture
    def property_instance(self):
        prop = QueryableProperty()
        prop.name = 'test_prop'
        return prop

    @pytest.fixture
    def model_instance(self):
        instance = ApplicationWithClassBasedProperties(name='Test')
        instance.__class__.dummy.counter = 0
        return instance

    def test_contribute_to_class(self, model_instance):
        prop = model_instance.__class__.dummy
        assert isinstance(prop, QueryableProperty)
        assert prop.name == 'dummy'
        assert prop.model is model_instance.__class__
        assert six.get_method_function(model_instance.reset_property) is reset_queryable_property
        # TODO: test that an existing method with the name reset_property will not be overridden

    def test_descriptor_get_class_attribute(self, model_instance):
        assert isinstance(model_instance.__class__.dummy, QueryableProperty)

    @pytest.mark.parametrize('cached, clear_cache, expected_values', [
        (False, False, [1, 2, 3, 4, 5]),  # The implementation of the dummy property returns increasing values ...
        (False, True, [1, 2, 3, 4, 5]),  # ... and cache clears shoudn't matter if caching was disabled all along
        (True, False, [1, 1, 1, 1, 1]),  # Caching is enabled (and never cleared); first value should always be returned
        (True, True, [1, 2, 3, 4, 5]),  # Cache is cleared every time; expect the same result like without caching
    ])
    def test_descriptor_get(self, monkeypatch, model_instance, cached, clear_cache, expected_values):
        monkeypatch.setattr(model_instance.__class__.dummy, 'cached', cached)
        for expected_value in expected_values:
            assert model_instance.dummy == expected_value
            if clear_cache:
                model_instance.reset_property('dummy')

    def test_descriptor_get_exception(self, monkeypatch, model_instance):
        monkeypatch.setattr(model_instance.__class__.dummy, 'get_value', None)
        with pytest.raises(AttributeError):
            model_instance.dummy

    @pytest.mark.parametrize('cached, setter_cache_behavior, values, expected_values', [
        # The setter cache behavior should not make a difference if a property
        # is not cached
        (False, CLEAR_CACHE, [0, 0, 0, 0, 0], [1, 2, 3, 4, 5]),
        (False, CACHE_VALUE, [0, 0, 0, 0, 0], [1, 2, 3, 4, 5]),
        (False, CACHE_RETURN_VALUE, [0, 0, 0, 0, 0], [1, 2, 3, 4, 5]),
        (False, DO_NOTHING, [0, 0, 0, 0, 0], [1, 2, 3, 4, 5]),
        (True, CLEAR_CACHE, [0, 0, 0, 0, 0], [1, 2, 3, 4, 5]),  # Getter always gets called again due to cache clear
        (True, CACHE_VALUE, [0, 0, 0, 0, 0], [0, 0, 0, 0, 0]),  # The raw value gets should get cached
        (True, CACHE_RETURN_VALUE, [0, 0, 0, 0, 0], [-1, -1, -1, -1, -1]),  # The return value should get cached
        (True, DO_NOTHING, [0, 0, 0, 0, 0], [1, 1, 1, 1, 1]),  # The first getter value should get and stay cached
    ])
    def test_descriptor_set(self, monkeypatch, model_instance, cached, setter_cache_behavior, values, expected_values):
        monkeypatch.setattr(model_instance.__class__.dummy, 'cached', cached)
        monkeypatch.setattr(model_instance.__class__.dummy, 'setter_cache_behavior', setter_cache_behavior)
        for value, expected_value in zip(values, expected_values):
            model_instance.dummy = value
            assert model_instance.dummy == expected_value

    @pytest.mark.parametrize('model', [VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties])
    def test_descriptor_set_exception(self, model):
        instance = model()
        with pytest.raises(AttributeError):
            instance.major_minor = '1.3'

    @pytest.mark.parametrize('value', [5, 'test', 1.337, None])
    def test_cache_methods(self, property_instance, model_instance, value):
        # Nothing should be cached initially
        assert not property_instance._has_cached_value(model_instance)
        with pytest.raises(KeyError):
            property_instance._get_cached_value(model_instance)
        # Clear calls should still work even if nothing is cached
        property_instance._clear_cached_value(model_instance)

        # Cache value, then try to use the other methods again
        property_instance._set_cached_value(model_instance, value)
        assert model_instance.__dict__['test_prop'] == value
        assert property_instance._has_cached_value(model_instance)
        assert property_instance._get_cached_value(model_instance) == value

        # Test if clearing the cached value works correctly
        property_instance._clear_cached_value(model_instance)
        assert not property_instance._has_cached_value(model_instance)
        with pytest.raises(KeyError):
            property_instance._get_cached_value(model_instance)

    @pytest.mark.parametrize('model', [VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties])
    def test_pickle_unpickle(self, model):
        serialized_prop = six.moves.cPickle.dumps(model.version)
        deserialized_prop = six.moves.cPickle.loads(serialized_prop)
        assert deserialized_prop is model.version

    def test_representations(self):
        string_representation = six.text_type(ApplicationWithClassBasedProperties.dummy)
        object_representation = repr(ApplicationWithClassBasedProperties.dummy)
        assert string_representation == 'tests.ApplicationWithClassBasedProperties.dummy'
        assert object_representation == '<DummyProperty: {}>'.format(string_representation)

    def test_invalid_property_name(self):
        with pytest.raises(QueryablePropertyError, match='must not contain the lookup separator'):
            type('BrokenModel', (Model,), {'dummy__dummy': DummyProperty(), '__module__': 'tests.models'})


class TestDecorators(object):

    ATTR_NAMES = ('get_value', 'set_value', 'get_filter', 'get_annotation', 'get_update_kwargs', 'cached',
                  'setter_cache_behavior', 'filter_requires_annotation', 'lookup_mappings', '__doc__')

    def assert_cloned_property(self, original, clone, changed_attrs):
        assert original is not clone
        for attr_name in self.ATTR_NAMES:
            value = getattr(clone, attr_name, None)
            assert value == changed_attrs.get(attr_name, getattr(original, attr_name, None))

    def decorate_function(self, func, decorator, decorator_kwargs=None):
        if decorator_kwargs is not None:
            decorator = decorator(**decorator_kwargs)
        return decorator(func)

    def test_extract_function(self):
        def func():
            pass
        cls_method = classmethod(func)

        prop = queryable_property()
        # Test both regular functions as well as class methods
        assert prop._extract_function(func) is func
        assert prop._extract_function(cls_method) is func

    @pytest.mark.parametrize('init_kwargs, clone_kwargs', [
        ({'getter': lambda: None}, {'set_value': lambda: None}),  # Set an additional attribute
        ({'getter': lambda: None}, {'get_value': lambda: 'test'}),  # Override an attribute
        ({'getter': lambda: None, 'cached': False},
         {'set_value': lambda: None, 'cached': True, '__doc__': 'my docstring'}),
    ])
    def test_clone(self, init_kwargs, clone_kwargs):
        prop = queryable_property(**init_kwargs)
        clone = prop._clone(**clone_kwargs)
        self.assert_cloned_property(prop, clone, clone_kwargs)

    @pytest.mark.parametrize('docstring, kwargs', [
        (None, None),  # Test @queryable_property
        (None, {}),  # Test @queryable_property() (without any arguments)
        ('my docstring', {}),  # Test if the docstring of the getter is used correctly
        (None, {'cached': True}),  # Keyword arguments
    ])
    def test_initializer(self, docstring, kwargs):
        def func():
            pass
        func.__doc__ = docstring
        prop = self.decorate_function(func, queryable_property, kwargs)
        assert prop.get_value is func
        assert prop.__doc__ == (docstring or None)
        if kwargs:
            for name, value in kwargs.items():
                assert getattr(prop, name) == value

    @pytest.mark.parametrize('old_getter, old_docstring, new_docstring, kwargs', [
        (None, None, None, None),
        (lambda: 'test', 'my old func', None, None),  # Test that the decorated function overrides an existing one
        (None, None, 'my new func', {'cached': True}),
        (lambda: 'test', 'my old func', 'my new func', {'cached': True}),
    ])
    def test_getter(self, old_getter, old_docstring, new_docstring, kwargs):
        original = queryable_property(old_getter)
        if old_docstring is not None:
            original.__doc__ = old_docstring

        def func():
            pass
        func.__doc__ = new_docstring

        clone = self.decorate_function(func, original.getter, kwargs)
        self.assert_cloned_property(original, clone,
                                    dict(kwargs or {}, get_value=func, __doc__=new_docstring or old_docstring))

    @pytest.mark.parametrize('old_setter, kwargs', [
        (None, None),
        (lambda: None, None),
        (None, {'setter_cache_behavior': DO_NOTHING}),
        (lambda: None, {'setter_cache_behavior': CACHE_VALUE}),
    ])
    def test_setter(self, old_setter, kwargs):
        original = queryable_property()
        original.set_value = old_setter

        def func():
            pass

        decorator_kwargs = kwargs and dict(kwargs)
        if decorator_kwargs:
            decorator_kwargs['cache_behavior'] = decorator_kwargs.pop('setter_cache_behavior')
        clone = self.decorate_function(func, original.setter, decorator_kwargs)
        self.assert_cloned_property(original, clone, dict(kwargs or {}, set_value=func))

    @pytest.mark.parametrize('initial_values, decorator_kwargs, expected_requires_annotation', [
        ({'get_filter': lambda: Q()}, {}, False),
        ({'get_filter': lambda: Q()}, None, False),
        # The following are the 9 cases for initial fra to decorator fra (each can be None, False or True)
        ({}, {}, False),
        ({'filter_requires_annotation': False}, {}, False),
        ({'filter_requires_annotation': True}, {}, True),
        ({}, {'requires_annotation': False}, False),
        ({'filter_requires_annotation': False}, {'requires_annotation': False}, False),
        ({'filter_requires_annotation': True}, {'requires_annotation': False}, False),
        ({}, {'requires_annotation': True}, True),
        ({'filter_requires_annotation': False}, {'requires_annotation': True}, True),
        ({'filter_requires_annotation': True}, {'requires_annotation': True}, True),
    ])
    def test_filter(self, initial_values, decorator_kwargs, expected_requires_annotation):
        original = queryable_property()
        original.__dict__.update(initial_values)

        def func():
            pass

        clone = self.decorate_function(func, original.filter, decorator_kwargs)
        assert not isinstance(clone, LookupFilterMixin)
        self.assert_cloned_property(original, clone,
                                    {'get_filter': func, 'filter_requires_annotation': expected_requires_annotation})

    def test_lookup_filters(self):
        original = queryable_property()
        original.model = Category
        original.name = 'test_property'
        get_filter_func = six.get_unbound_function(LookupFilterMixin.get_filter)

        def func1(cls, lookup, value):
            return 1

        clone1 = self.decorate_function(func1, original.filter, {'lookups': ['lt', 'gt']})
        assert isinstance(clone1, LookupFilterMixin)
        self.assert_cloned_property(original, clone1, {
            'lookup_mappings': {'lt': func1, 'gt': func1},
            'get_filter': six.create_bound_method(get_filter_func, clone1),
        })
        assert clone1.get_filter(None, 'lt', None) == 1
        assert clone1.get_filter(None, 'gt', None) == 1
        with pytest.raises(QueryablePropertyError):
            clone1.get_filter(None, 'in', None)

        def func2(cls, lookup, value):
            return 2

        clone2 = self.decorate_function(func2, clone1.filter, {'lookups': ['lt', 'lte'], 'requires_annotation': True})
        assert isinstance(clone2, LookupFilterMixin)
        self.assert_cloned_property(clone1, clone2, {
            'lookup_mappings': {'lt': func2, 'lte': func2, 'gt': func1},
            'get_filter': six.create_bound_method(get_filter_func, clone2),
            'filter_requires_annotation': True,  # Should be overridable on every call.
        })
        assert clone2.get_filter(None, 'lt', None) == 2
        assert clone2.get_filter(None, 'lte', None) == 2
        assert clone2.get_filter(None, 'gt', None) == 1
        with pytest.raises(QueryablePropertyError):
            clone2.get_filter(None, 'in', None)

        def func3(cls, lookup, value):
            return 3

        clone3 = self.decorate_function(func3, clone2.filter, {'lookups': ['exact'], 'requires_annotation': False})
        assert isinstance(clone3, LookupFilterMixin)
        self.assert_cloned_property(clone2, clone3, {
            'lookup_mappings': {'lt': func2, 'lte': func2, 'gt': func1, 'exact': func3},
            'get_filter': six.create_bound_method(get_filter_func, clone3),
            'filter_requires_annotation': False,  # Should be overridable on every call.
        })
        assert clone3.get_filter(None, 'lt', None) == 2
        assert clone3.get_filter(None, 'lte', None) == 2
        assert clone3.get_filter(None, 'gt', None) == 1
        assert clone3.get_filter(None, 'exact', None) == 3
        with pytest.raises(QueryablePropertyError):
            clone3.get_filter(None, 'in', None)

    def test_boolean_filter(self):
        original = queryable_property()
        original.model = Category
        original.name = 'test_property'

        def func(cls):
            return Q(some_field=5)

        clone = self.decorate_function(func, original.filter, {'boolean': True})
        assert isinstance(clone, LookupFilterMixin)
        positive_condition = clone.get_filter(None, 'exact', True)
        negative_condition = clone.get_filter(None, 'exact', False)
        if DJANGO_VERSION < (1, 6):
            # In very old Django versions, negating adds another layer.
            negative_condition = negative_condition.children[0]
        assert positive_condition.children == negative_condition.children == [('some_field', 5)]
        assert positive_condition.negated is False
        assert negative_condition.negated is True
        with pytest.raises(QueryablePropertyError):
            clone.get_filter(None, 'lt', None)

    def test_lookup_boolean_exception(self):
        prop = queryable_property()

        def func():
            pass

        with pytest.raises(QueryablePropertyError):
            self.decorate_function(func, prop.filter, {'lookups': ['lt', 'lte'], 'boolean': True})

    @pytest.mark.parametrize('initial_values, expected_requires_annotation', [
        ({'get_annotation': lambda: F('dummy')}, True),
        ({}, True),
        ({'filter_requires_annotation': False}, False),
        ({'filter_requires_annotation': True}, True),
        ({'get_filter': lambda: Q()}, True),
        ({'get_filter': lambda: Q(), 'filter_requires_annotation': False}, False),
        ({'get_filter': lambda: Q(), 'filter_requires_annotation': True}, True),
    ])
    def test_annotater(self, initial_values, expected_requires_annotation):
        original = queryable_property()
        original.__dict__.update(initial_values)

        def func():
            pass

        prop = self.decorate_function(func, original.annotater)
        assert isinstance(prop, AnnotationMixin)
        self.assert_cloned_property(original, prop, {
            'get_annotation': func,
            'filter_requires_annotation': expected_requires_annotation,
            'get_filter': initial_values.get(
                'get_filter', six.create_bound_method(six.get_unbound_function(AnnotationMixin.get_filter), prop)),
        })

    @pytest.mark.parametrize('old_updater', [None, lambda: {}])
    def test_updater(self, old_updater):
        original = queryable_property()
        original.get_update_kwargs = old_updater

        def func():
            pass

        clone = self.decorate_function(classmethod(func), original.updater)
        self.assert_cloned_property(original, clone, {'get_update_kwargs': func})



# encoding: utf-8
"""Write Django model properties that can be used in database queries."""

from __future__ import unicode_literals

__version__ = '1.4.0'
__author__ = 'Marcus Klöpfel'
__copyright__ = 'Copyright 2020, Marcus Klöpfel'
__license__ = 'BSD'
__maintainer__ = 'Marcus Klöpfel'
__email__ = 'marcus.kloepfel@gmail.com'
__status__ = 'Production/Stable'



# encoding: utf-8

import pytest

import six
from django import VERSION as DJANGO_VERSION
from django.db import models

from ..models import (ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties,
                      CategoryWithClassBasedProperties, CategoryWithDecoratorBasedProperties,
                      VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties)

pytestmark = [pytest.mark.django_db, pytest.mark.usefixtures('versions')]


class TestFilterWithoutAnnotations(object):

    @pytest.mark.parametrize('model, filters, expected_count, expected_major_minor', [
        # Test that filter that don't involve queryable properties still work
        (VersionWithClassBasedProperties, models.Q(major=2, minor=0), 2, '2.0'),
        (VersionWithDecoratorBasedProperties, models.Q(major=2, minor=0), 2, '2.0'),
        (VersionWithClassBasedProperties, models.Q(minor=2) | models.Q(patch=3), 2, '1.2'),
        (VersionWithDecoratorBasedProperties, models.Q(minor=2) | models.Q(patch=3), 2, '1.2'),
        # All querysets are expected to return objects with the same
        # major_minor value (major_minor parameter).
        (VersionWithClassBasedProperties, models.Q(major_minor='1.2'), 2, '1.2'),
        (VersionWithDecoratorBasedProperties, models.Q(major_minor='1.2'), 2, '1.2'),
        # Also test that using non-property filters still work and can be used
        # together with filters for queryable properties
        (VersionWithClassBasedProperties, models.Q(major_minor='1.3') & models.Q(major=1), 4, '1.3'),
        (VersionWithDecoratorBasedProperties, models.Q(major_minor='1.3') & models.Q(major=1), 4, '1.3'),
        (VersionWithClassBasedProperties, models.Q(major_minor='1.3') | models.Q(patch=1), 4, '1.3'),
        (VersionWithDecoratorBasedProperties, models.Q(major_minor='1.3') | models.Q(patch=1), 4, '1.3'),
        # Also test nested filters
        (VersionWithClassBasedProperties, (models.Q(major_minor='2.0') | models.Q(patch=0)) & models.Q(minor=0),
         2, '2.0'),
        (VersionWithDecoratorBasedProperties, (models.Q(major_minor='2.0') | models.Q(patch=0)) & models.Q(minor=0),
         2, '2.0'),
    ])
    def test_simple_filter(self, model, filters, expected_count, expected_major_minor):
        queryset = model.objects.filter(filters)
        assert queryset.count() == len(queryset) == expected_count
        assert all(obj.major_minor == expected_major_minor for obj in queryset)

    @pytest.mark.parametrize('model, property_name, value, expected_count, record_checker', [
        # Filtering the 'version' property is also based on filtering the
        # 'major_minor' property, so this test also tests properties that build
        # on each other.
        (VersionWithClassBasedProperties, 'version', '1.2.3', 2, lambda obj: obj.version == '1.2.3'),
        (VersionWithDecoratorBasedProperties, 'version', '1.2.3', 2, lambda obj: obj.version == '1.2.3'),
        # Filters across relations
        (ApplicationWithClassBasedProperties, 'versions__version', '1.2.3', 2,
         lambda obj: obj.versions.filter(version='1.2.3').exists()),
        (ApplicationWithDecoratorBasedProperties, 'versions__version', '1.2.3', 2,
         lambda obj: obj.versions.filter(version='1.2.3').exists()),
    ])
    def test_filter_without_required_annotation(self, model, property_name, value, expected_count, record_checker):
        queryset = model.objects.filter(**{property_name: value})
        assert property_name not in queryset.query.annotations
        assert queryset.count() == len(queryset) == expected_count
        assert all(record_checker(obj) for obj in queryset)

    @pytest.mark.parametrize('model, filters, expected_remaining_count', [
        (VersionWithClassBasedProperties, models.Q(major_minor='1.3'), 4),
        (VersionWithDecoratorBasedProperties, models.Q(major_minor='1.3'), 4),
        (VersionWithClassBasedProperties, models.Q(version='2.0.0'), 6),
        (VersionWithDecoratorBasedProperties, models.Q(major_minor='2.0.0'), 6),
        (VersionWithClassBasedProperties, models.Q(major_minor='1.3', patch=1), 6),
        (VersionWithDecoratorBasedProperties, models.Q(major_minor='1.3', patch=1), 6),
        (VersionWithClassBasedProperties, models.Q(major_minor='1.3') | models.Q(version='2.0.0'), 2),
        (VersionWithDecoratorBasedProperties, models.Q(major_minor='1.3') | models.Q(version='2.0.0'), 2),
        # Filters across relations
        (ApplicationWithClassBasedProperties, models.Q(versions__version='2.0.0'), 0),
        (ApplicationWithDecoratorBasedProperties, models.Q(versions__version='2.0.0'), 0),
    ])
    def test_delete_query(self, model, filters, expected_remaining_count):
        model.objects.filter(filters).delete()
        assert model.objects.count() == expected_remaining_count

    @pytest.mark.parametrize('model', [VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties])
    def test_boolean_filter(self, model):
        v2_objects = model.objects.filter(is_version_2=True)
        non_v2_objects = model.objects.filter(is_version_2=False)
        assert len(v2_objects) == 2
        assert all(version.is_version_2 for version in v2_objects)
        assert len(non_v2_objects) == 6
        assert all(not version.is_version_2 for version in non_v2_objects)

    @pytest.mark.skipif(DJANGO_VERSION < (1, 8), reason="Expression-based annotations didn't exist before Django 1.8")
    @pytest.mark.parametrize('model, property_name, condition', [
        (VersionWithClassBasedProperties, 'major_minor', models.Q(major_minor='1.3')),
        (VersionWithDecoratorBasedProperties, 'major_minor', models.Q(major_minor='1.3')),
        (VersionWithClassBasedProperties, 'version', models.Q(version='1.3.0') | models.Q(version='1.3.1')),
        (VersionWithDecoratorBasedProperties, 'version', models.Q(version='1.3.0') | models.Q(version='1.3.1')),
    ])
    def test_filter_in_case_expression(self, model, property_name, condition):
        queryset = model.objects.annotate(is_13=models.Case(
            models.When(condition, then=1),
            default=0,
            output_field=models.IntegerField()
        ))
        assert property_name not in queryset.query.annotations
        assert all(bool(version.is_13) is (version.major_minor == '1.3') for version in queryset)

    @pytest.mark.skipif(DJANGO_VERSION < (2, 0), reason="Per-aggregate filters didn't exist before Django 2.0")
    @pytest.mark.parametrize('model', [ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties])
    def test_filter_in_aggregate(self, model):
        queryset = model.objects.annotate(
            num_13=models.Count('versions', filter=models.Q(versions__major_minor='1.3')),
            num_200=models.Count('versions', filter=models.Q(versions__version='2.0.0'))
        )
        assert 'major_minor' not in queryset.query.annotations
        assert 'version' not in queryset.query.annotations
        for app in queryset:
            assert app.num_13 == 2
            assert app.num_200 == 1

    @pytest.mark.skipif(DJANGO_VERSION < (3, 0), reason="Conditional filters didn't exist before Django 3.0")
    @pytest.mark.parametrize('application_model, version_model', [
        (ApplicationWithClassBasedProperties, VersionWithClassBasedProperties),
        (ApplicationWithDecoratorBasedProperties, VersionWithDecoratorBasedProperties),
    ])
    def test_conditional_filter(self, application_model, version_model):
        version_model.objects.filter(major=2).first().delete()
        subquery = version_model.objects.filter(application_id=models.OuterRef('pk'), version='2.0.0')
        applications = application_model.objects.filter(models.Exists(subquery))
        assert len(applications) == 1
        assert applications[0].versions.filter(version='2.0.0').exists()


class TestFilterWithAggregateAnnotation(object):

    @pytest.mark.parametrize('model, property_name, filters, expected_count', [
        (ApplicationWithClassBasedProperties, 'version_count', models.Q(version_count__gt=3), 2),
        (ApplicationWithClassBasedProperties, 'version_count', models.Q(version_count=4, name__contains='cool'), 1),
        (ApplicationWithClassBasedProperties, 'version_count',
         models.Q(version_count=4) | models.Q(name__contains='cool'), 2),
        (ApplicationWithClassBasedProperties, 'version_count', models.Q(version_count__gt=3, major_sum__gt=5), 0),
        (ApplicationWithClassBasedProperties, 'version_count',
         models.Q(version_count__gt=3) | models.Q(major_sum__gt=5), 2),
        (ApplicationWithDecoratorBasedProperties, 'version_count', models.Q(version_count__gt=3), 2),
        (ApplicationWithDecoratorBasedProperties, 'version_count',
         models.Q(version_count=4, name__contains='cool'), 1),
        (ApplicationWithDecoratorBasedProperties, 'version_count',
         models.Q(version_count=4) | models.Q(name__contains='cool'), 2),
        (ApplicationWithDecoratorBasedProperties, 'version_count', models.Q(version_count__gt=3, major_sum__gt=5), 0),
        (ApplicationWithDecoratorBasedProperties, 'version_count',
         models.Q(version_count__gt=3) | models.Q(major_sum__gt=5), 2),
        # Filters across relations
        # A query containing an aggregate across a relation is still only
        # grouped by fields of the query's model, so in this case the version
        # count is the total number of application versions per category.
        (CategoryWithClassBasedProperties, 'applications__version_count', models.Q(applications__version_count=4), 1),
        (CategoryWithDecoratorBasedProperties, 'applications__version_count',
         models.Q(applications__version_count=4), 1),
        (CategoryWithClassBasedProperties, 'applications__version_count', models.Q(applications__version_count=8), 1),
        (CategoryWithDecoratorBasedProperties, 'applications__version_count',
         models.Q(applications__version_count=8), 1),
        (CategoryWithClassBasedProperties, 'applications__version_count',
         models.Q(applications__version_count=4, name__contains='Windows'), 1),
        (CategoryWithDecoratorBasedProperties, 'applications__version_count',
         models.Q(applications__version_count=4, name__contains='Windows'), 1),
        (CategoryWithClassBasedProperties, 'applications__version_count',
         models.Q(applications__version_count=8, name__contains='Windows'), 0),
        (CategoryWithDecoratorBasedProperties, 'applications__version_count',
         models.Q(applications__version_count=8, name__contains='Windows'), 0),
    ])
    def test_select_query(self, model, property_name, filters, expected_count):
        queryset = model.objects.filter(filters)
        assert property_name in queryset.query.annotations
        assert queryset.count() == len(queryset) == expected_count
        if '__' not in property_name:
            # Check that a property annotation used implicitly by a filter does
            # not lead to a selection of the property annotation.
            prop = getattr(model, property_name)
            assert all(not prop._has_cached_value(app) for app in queryset)

    @pytest.mark.parametrize('model, filters, expected_remaining_count', [
        (ApplicationWithClassBasedProperties, models.Q(version_count=4, name__contains='cool'), 1),
        (ApplicationWithDecoratorBasedProperties, models.Q(version_count=4, name__contains='cool'), 1),
        # Filters across relations
        (CategoryWithClassBasedProperties, models.Q(applications__version_count=4, name__contains='Windows'), 1),
        (CategoryWithDecoratorBasedProperties, models.Q(applications__version_count=4, name__contains='Windows'), 1),
    ])
    def test_delete_query(self, model, filters, expected_remaining_count):
        model.objects.filter(filters).delete()
        assert model.objects.count() == expected_remaining_count

    @pytest.mark.parametrize('model', [ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties])
    def test_filter_implementation_used_despite_present_annotation(self, monkeypatch, model):
        # Patch the property to have a filter that is always True, then use a
        # condition that would be False without the patch.
        monkeypatch.setattr(model.version_count, 'get_filter', lambda cls, lookup, value: models.Q(pk__gt=0))
        queryset = model.objects.select_properties('version_count').filter(version_count__gt=5)
        assert '"id" > 0' in six.text_type(queryset.query)
        assert queryset.count() == len(queryset) == 2


@pytest.mark.skipif(DJANGO_VERSION < (1, 8), reason="Expression-based annotations didn't exist before Django 1.8")
class TestFilterWithExpressionAnnotation(object):

    @pytest.mark.parametrize('model, property_name, filters, expected_count, expected_distinct_count, record_checker', [
        (VersionWithClassBasedProperties, 'changes_or_default', models.Q(changes_or_default='(No data)'), 6, 6,
         lambda obj: obj.changes_or_default == '(No data)'),
        (VersionWithDecoratorBasedProperties, 'changes_or_default', models.Q(changes_or_default='(No data)'), 6, 6,
         lambda obj: obj.changes_or_default == '(No data)'),
        (VersionWithClassBasedProperties, 'changes_or_default',
         models.Q(changes_or_default='(No data)') | models.Q(major=2), 8, 8,
         lambda obj: obj.changes_or_default == '(No data)' or obj.major == 2),
        (VersionWithDecoratorBasedProperties, 'changes_or_default',
         models.Q(changes_or_default='(No data)') | models.Q(major=2), 8, 8,
         lambda obj: obj.changes_or_default == '(No data)' or obj.major == 2),
        # Filters across relations
        (ApplicationWithClassBasedProperties, 'versions__changes_or_default',
         models.Q(versions__changes_or_default='(No data)'), 6, 2,
         lambda obj: obj.versions.filter(changes_or_default='(No data)').exists()),
        (ApplicationWithDecoratorBasedProperties, 'versions__changes_or_default',
         models.Q(versions__changes_or_default='(No data)'), 6, 2,
         lambda obj: obj.versions.filter(changes_or_default='(No data)').exists()),
        (ApplicationWithClassBasedProperties, 'versions__changes_or_default',
         models.Q(versions__minor=3) & models.Q(versions__changes_or_default='(No data)'), 4, 2,
         lambda obj: obj.versions.filter(changes_or_default='(No data)').exists()),
        (ApplicationWithDecoratorBasedProperties, 'versions__changes_or_default',
         models.Q(versions__minor=3) & models.Q(versions__changes_or_default='(No data)'), 4, 2,
         lambda obj: obj.versions.filter(changes_or_default='(No data)').exists()),
        # Filters across relations with dependencies across other relations
        (CategoryWithClassBasedProperties, 'applications__lowered_version_changes',
         models.Q(applications__lowered_version_changes='amazing new features'), 3, 2, None),
        (CategoryWithDecoratorBasedProperties, 'applications__lowered_version_changes',
         models.Q(applications__lowered_version_changes='amazing new features'), 3, 2, None),
    ])
    def test_select_query(self, model, property_name, filters, expected_count, expected_distinct_count, record_checker):
        queryset = model.objects.filter(filters)
        assert property_name in queryset.query.annotations
        assert queryset.distinct().count() == expected_distinct_count
        assert len(queryset) == expected_count
        if record_checker:
            assert all(record_checker(obj) for obj in queryset)
        if '__' not in property_name:
            # Check that a property annotation used implicitly by a filter does
            # not lead to a selection of the property annotation.
            prop = getattr(model, property_name)
            assert all(not prop._has_cached_value(version) for version in queryset)

    @pytest.mark.parametrize('model, filters, expected_remaining_count', [
        (VersionWithClassBasedProperties, models.Q(changes_or_default='(No data)', major=1), 2),
        (VersionWithDecoratorBasedProperties, models.Q(changes_or_default='(No data)', major=1), 2),
        # Filters across relations
        (ApplicationWithClassBasedProperties, models.Q(versions__changes_or_default='(No data)'), 0),
        (ApplicationWithDecoratorBasedProperties, models.Q(versions__changes_or_default='(No data)'), 0),
    ])
    def test_delete_query(self, model, filters, expected_remaining_count):
        model.objects.filter(filters).delete()
        assert model.objects.count() == expected_remaining_count

    @pytest.mark.parametrize('model', [VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties])
    def test_filter_implementation_used_despite_present_annotation(self, model):
        queryset = model.objects.select_properties('version').filter(version='2.0.0')
        pseudo_sql = six.text_type(queryset.query)
        assert '"major" = 2' in pseudo_sql
        assert '"minor" = 0' in pseudo_sql
        assert '"patch" = 0' in pseudo_sql
        assert queryset.count() == len(queryset) == 2


@pytest.mark.skipif(DJANGO_VERSION < (1, 11), reason="Explicit subqueries didn't exist before Django 1.11")
class TestFilterWithSubqueryAnnotation(object):

    @pytest.mark.parametrize('model, property_name, filters, expected_count, record_checker', [
        (ApplicationWithClassBasedProperties, 'highest_version', models.Q(highest_version='2.0.0'), 1,
         lambda obj: obj.versions.filter(major=2, minor=0, patch=0).exists()),
        (ApplicationWithDecoratorBasedProperties, 'highest_version', models.Q(highest_version='2.0.0'), 1,
         lambda obj: obj.versions.filter(major=2, minor=0, patch=0).exists()),
        (ApplicationWithClassBasedProperties, 'highest_version',
         models.Q(highest_version='2.0.0') | models.Q(name__startswith='Another'), 1,
         lambda obj: obj.versions.filter(major=2, minor=0, patch=0).exists()),
        (ApplicationWithDecoratorBasedProperties, 'highest_version',
         models.Q(highest_version='2.0.0') | models.Q(name__startswith='Another'), 1,
         lambda obj: obj.versions.filter(major=2, minor=0, patch=0).exists()),
        # Filters across relations
        (VersionWithClassBasedProperties, 'application__highest_version',
         models.Q(application__highest_version='2.0.0'), 4, lambda obj: obj.application.highest_version == '2.0.0'),
        (VersionWithDecoratorBasedProperties, 'application__highest_version',
         models.Q(application__highest_version='2.0.0'), 4, lambda obj: obj.application.highest_version == '2.0.0'),
        (VersionWithClassBasedProperties, 'application__highest_version',
         models.Q(application__name__startswith='Another') & models.Q(application__highest_version='2.0.0'), 4,
         lambda obj: obj.application.highest_version == '2.0.0'),
        (VersionWithDecoratorBasedProperties, 'application__highest_version',
         models.Q(application__name__startswith='Another') & models.Q(application__highest_version='2.0.0'), 4,
         lambda obj: obj.application.highest_version == '2.0.0'),
    ])
    def test_select_query(self, model, property_name, filters, expected_count, record_checker):
        version_model = (VersionWithClassBasedProperties if 'ClassBased' in model.__name__
                         else VersionWithDecoratorBasedProperties)
        # Delete one version to create distinct constellations
        version_model.objects.get(version='2.0.0', application__name__contains='cool').delete()
        queryset = model.objects.filter(filters)
        assert property_name in queryset.query.annotations
        assert queryset.count() == len(queryset) == expected_count
        if '__' not in property_name:
            # Check that a property annotation used implicitly by a filter does
            # not lead to a selection of the property annotation.
            prop = getattr(model, property_name)
            assert all(not prop._has_cached_value(version) for version in queryset)

    @pytest.mark.parametrize('model, filters, expected_remaining_count', [
        (ApplicationWithClassBasedProperties, models.Q(highest_version='2.0.0', name__contains='cool'), 1),
        (ApplicationWithDecoratorBasedProperties, models.Q(highest_version='2.0.0', name__contains='cool'), 1),
        # Filters across relations
        (CategoryWithClassBasedProperties, models.Q(applications__highest_version='2.0.0'), 0),
        (CategoryWithDecoratorBasedProperties, models.Q(applications__highest_version='2.0.0'), 0),
    ])
    def test_delete_query(self, model, filters, expected_remaining_count):
        model.objects.filter(filters).delete()
        assert model.objects.count() == expected_remaining_count



# encoding: utf-8

import pytest

from django import VERSION as DJANGO_VERSION
from django.db import models

from ..models import VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties

pytestmark = [pytest.mark.django_db, pytest.mark.usefixtures('versions')]


@pytest.mark.parametrize('model', [VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties])
def test_simple_update(model):
    queryset = model.objects.filter(major_minor='2.0')
    pks = list(queryset.values_list('pk', flat=True))
    assert queryset.update(major_minor='42.42') == len(pks)
    for pk in pks:
        version = model.objects.get(pk=pk)  # Reload from DB
        assert version.major_minor == '42.42'


@pytest.mark.parametrize('model, update_kwargs', [
    (VersionWithClassBasedProperties, {'version': '1.3.37'}),
    (VersionWithDecoratorBasedProperties, {'version': '1.3.37'}),
    # Also test that setting the same field(s) via multiple queryable
    # properties works as long as they try to set the same values
    (VersionWithClassBasedProperties, {'version': '1.3.37', 'major_minor': '1.3'}),
    (VersionWithDecoratorBasedProperties, {'version': '1.3.37', 'major_minor': '1.3'}),
])
def test_update_based_on_other_property(model, update_kwargs):
    queryset = model.objects.filter(version='1.3.1')
    pks = list(queryset.values_list('pk', flat=True))
    assert queryset.update(**update_kwargs) == len(pks)
    for pk in pks:
        version = model.objects.get(pk=pk)  # Reload from DB
        assert version.version == update_kwargs['version']


@pytest.mark.skipif(DJANGO_VERSION < (1, 8), reason="Conditional expressions didn't exist before Django 1.8")
@pytest.mark.parametrize('model', [VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties])
def test_update_using_conditional_expression(model):
    assert model.objects.filter(major=1, changes__isnull=True).count() == 6
    model.objects.filter(major=1).update(changes=models.Case(
        models.When(version='1.2.3', then=models.Value('1.2.3 changes')),
        models.When(version='1.3.0', then=models.Value('1.3.0 changes')),
        default=models.Value('1.3.1 changes')
    ))
    for version in ('1.2.3', '1.3.0', '1.3.1'):
        assert model.objects.filter(major=1, version=version, changes='{} changes'.format(version)).count() == 2



# encoding: utf-8
"""A stable import interface for Django classes that were moved in between versions and compatibility constants."""

try:  # pragma: no cover
    from contextlib import nullcontext  # noqa: F401
except ImportError:  # pragma: no cover
    from contextlib import contextmanager

    @contextmanager
    def nullcontext(enter_result=None):
        yield enter_result

import six
from django.db.models.query import QuerySet
from django.db.models.sql.query import Query

try:  # pragma: no cover
    from django.db.models.constants import LOOKUP_SEP  # noqa: F401
except ImportError:  # pragma: no cover
    from django.db.models.sql.constants import LOOKUP_SEP  # noqa: F401

try:  # pragma: no cover
    from django.db.models.query import ModelIterable  # noqa: F401
    ValuesQuerySet = None
except ImportError:  # pragma: no cover
    from django.db.models.query import ValuesQuerySet  # noqa: F401
    ModelIterable = None

# A dictionary mapping names of build_filter/add_filter keyword arguments to
# keyword arguments for an _add_q/add_q call. It contains kwargs names for
# all Django versions (some do not use all of these). If a keyword argument
# is not part of this dictionary, it will not be passed through.
BUILD_FILTER_TO_ADD_Q_KWARGS_MAP = {
    'can_reuse': 'used_aliases',
    'branch_negated': 'branch_negated',
    'current_negated': 'current_negated',
    'allow_joins': 'allow_joins',
    'split_subq': 'split_subq',
    'force_having': 'force_having',
}

# Very old django versions (<1.6) had different names for the methods
# containing the build_filter and _add_q logic, which are needed as the core
# for filters based on queryable properties.
BUILD_FILTER_METHOD_NAME = 'build_filter'
ADD_Q_METHOD_NAME = '_add_q'
if not hasattr(Query, 'build_filter'):  # pragma: no cover
    BUILD_FILTER_METHOD_NAME = 'add_filter'
    ADD_Q_METHOD_NAME = 'add_q'

# Old Django versions (<1.9) had a method to check if filter conditions need
# to be put into the HAVING clause instead of the WHERE clause. To get
# queryable properties based on aggregates to work, these methods must be
# intercepted if present.
NEED_HAVING_METHOD_NAME = None
if hasattr(Query, 'need_having'):  # pragma: no cover
    NEED_HAVING_METHOD_NAME = 'need_having'
elif hasattr(Query, 'need_force_having'):  # pragma: no cover
    NEED_HAVING_METHOD_NAME = 'need_force_having'

# The annotation-related attributes of Query objects had "aggregate" in their
# name instead of "annotation" in old django versions (<1.8), because
# annotations were strongly tied to aggregates.
ANNOTATION_TO_AGGREGATE_ATTRIBUTES_MAP = {}
ANNOTATION_SELECT_CACHE_NAME = '_annotation_select_cache'
if not hasattr(Query, 'annotation_select'):  # pragma: no cover
    ANNOTATION_TO_AGGREGATE_ATTRIBUTES_MAP = {
        'add_annotation': 'add_aggregate',
        '_annotations': '_aggregates',
        'annotations': 'aggregates',
        '_annotation_select_cache': '_aggregate_select_cache',
        'annotation_select': 'aggregate_select',
        'annotation_select_mask': 'aggregate_select_mask',
        'set_annotation_mask': 'set_aggregate_mask',
    }
    ANNOTATION_SELECT_CACHE_NAME = '_aggregate_select_cache'

# Recent Django versions (>=2.0) have separate methods for cloning and chaining
# while older versions only have the clone method.
QUERYSET_CHAIN_METHOD_NAME = '_chain' if hasattr(QuerySet, '_chain') else '_clone'
QUERY_CHAIN_METHOD_NAME = 'chain' if hasattr(Query, 'chain') else 'clone'


def convert_build_filter_to_add_q_kwargs(**build_filter_kwargs):
    """
    Transform the keyword arguments of a :meth:`Query.build_filter` call into
    keyword arguments for an appropriate :meth:`Query._add_q` call (or their
    respective counterparts in older Django versions).

    :param build_filter_kwargs: The keyword arguments passed to
                                :meth:`Query.build_filter`.
    :return: The keywords argument to use for :meth:`Query._add_q`.
    :rtype: dict
    """
    return {BUILD_FILTER_TO_ADD_Q_KWARGS_MAP[key]: value for key, value in six.iteritems(build_filter_kwargs)
            if key in BUILD_FILTER_TO_ADD_Q_KWARGS_MAP}


def chain_queryset(queryset, *args, **kwargs):
    """
    Create a copy of the given queryset to chain a new queryset method call by
    calling the appropriate chain/clone method for the current Django version.

    :param QuerySet queryset: The queryset to chain.
    :param args: Positional arguments passed through to the method call.
    :param kwargs: Keyword arguments passed through to the method call.
    :return: A copy of given queryset.
    :rtype: QuerySet
    """
    method = getattr(queryset, QUERYSET_CHAIN_METHOD_NAME)
    return method(*args, **kwargs)


def chain_query(query, *args, **kwargs):
    """
    Create a copy of the given query to chain a new query method call by
    calling the appropriate chain/clone method for the current Django version.

    :param Query query: The query to chain.
    :param args: Positional arguments passed through to the method call.
    :param kwargs: Keyword arguments passed through to the method call.
    :return: A copy of given query.
    :rtype: Query
    """
    method = getattr(query, QUERY_CHAIN_METHOD_NAME)
    return method(*args, **kwargs)


def contains_aggregate(annotation):
    """
    Check if the given annotation contains an aggregate.

    :param annotation: The annotation to check.
    :return: True if the annotation contains an aggregate; otherwise False.
    :rtype: bool
    """
    # While annotations can mark themselves as containing an aggregate via the
    # contains_aggregate attribute in recent Django versions, this was not the
    # case in old versions. Annotations were strongly tied to aggregates in
    # these versions though, so an aggregate is always assumed in this case.
    return getattr(annotation, 'contains_aggregate', bool(ANNOTATION_TO_AGGREGATE_ATTRIBUTES_MAP))


def get_related_model(model, relation_field_name):
    """
    Get the related model of the (presumed) relation field with the given name
    on the given model.

    :param type model: The model class to inspect the field on.
    :param str relation_field_name: The field name of the (presumed) relation
                                    field.
    :return: The model class reached via the relation field or None if the
             field is not actually a relation field.
    :rtype: type | None
    """
    if hasattr(model._meta, 'get_field_by_name'):  # pragma: no cover
        # Older Django versions (<1.8) only allowed to find reverse relation
        # objects as well as fields via the get_field_by_name method, which
        # doesn't exist in recent versions anymore.
        field_or_rel, _, direct, _ = model._meta.get_field_by_name(relation_field_name)
        # Unlike in recent Django versions, the reverse relation objects and
        # fields also didn't provide the same attributes, which is why they
        # need to be treated differently.
        if not direct:  # direct=False means a reverse relation object
            return field_or_rel.field.model
        return field_or_rel.rel and field_or_rel.rel.to
    return model._meta.get_field(relation_field_name).related_model



# encoding: utf-8

from collections import namedtuple
from contextlib import contextmanager
from functools import partial

import six
from django.utils.tree import Node

from .compat import (
    ADD_Q_METHOD_NAME, ANNOTATION_TO_AGGREGATE_ATTRIBUTES_MAP, BUILD_FILTER_METHOD_NAME, contains_aggregate,
    convert_build_filter_to_add_q_kwargs, get_related_model, LOOKUP_SEP, NEED_HAVING_METHOD_NAME, nullcontext,
    QUERY_CHAIN_METHOD_NAME
)
from .exceptions import FieldDoesNotExist, QueryablePropertyDoesNotExist, QueryablePropertyError
from .utils import get_queryable_property, InjectableMixin, TreeNodeProcessor


class QueryablePropertyReference(namedtuple('QueryablePropertyReference', 'property model relation_path')):
    """
    A reference to a queryable property that also holds the path to reach the
    property across relations.
    """
    __slots__ = ()

    @property
    def full_path(self):
        """
        Return the full path to the queryable property (including the relation
        prefix) in the query filter format.

        :return: The full path to the queryable property.
        :rtype: str
        """
        if not self.relation_path:
            return self.property.name
        return LOOKUP_SEP.join(self.relation_path + (self.property.name,))

    def get_filter(self, lookups, value):
        """
        A wrapper for the get_filter method of the property this reference
        points to. It checks if the property actually supports filtering and
        applies the relation path (if any) to the returned Q object.

        :param collections.Sequence[str] lookups: The lookups/transforms to use
                                                  for the filter.
        :param value: The value passed to the filter condition.
        :return: A Q object to filter using this property.
        :rtype: django.db.models.Q
        """
        if not self.property.get_filter:
            raise QueryablePropertyError('Queryable property "{}" is supposed to be used as a filter but does not '
                                         'implement filtering.'.format(self.property))

        # Use the model stored on this reference instead of the one on the
        # property since the query may be happening from a subclass of the
        # model the property is defined on.
        q_obj = self.property.get_filter(self.model, LOOKUP_SEP.join(lookups) or 'exact', value)
        if self.relation_path:
            # If the resolved property belongs to a related model, all actual
            # conditions in the returned Q object must be modified to use the
            # current relation path as prefix.
            def prefix_condition(item):
                return LOOKUP_SEP.join(self.relation_path + (item[0],)), item[1]
            q_obj = TreeNodeProcessor(q_obj).modify_leaves(prefix_condition)
        return q_obj

    def get_annotation(self):
        """
        A wrapper for the get_annotation method of the property this reference
        points to. It checks if the property actually supports annotation
        creation performs the internal call with the correct model class.

        :return: An annotation object.
        """
        if not self.property.get_annotation:
            raise QueryablePropertyError('Queryable property "{}" needs to be added as annotation but does not '
                                         'implement annotation creation.'.format(self.property))
        # Use the model stored on this reference instead of the one on the
        # property since the query may be happening from a subclass of the
        # model the property is defined on.
        return self.property.get_annotation(self.model)


class QueryablePropertiesQueryMixin(InjectableMixin):
    """
    A mixin for :class:`django.db.models.sql.Query` objects that extends the
    original Django objects to deal with queryable properties, e.g. managing
    used properties or automatically adding required properties as annotations.
    """

    def __getattr__(self, name):  # pragma: no cover
        # Redirect some attribute accesses for older Django versions (where
        # annotations were tied to aggregations, hence "aggregation" in the
        # names instead of "annotation").
        if name in ANNOTATION_TO_AGGREGATE_ATTRIBUTES_MAP:
            return getattr(self, ANNOTATION_TO_AGGREGATE_ATTRIBUTES_MAP[name])
        raise AttributeError()

    def init_injected_attrs(self):
        # Stores references to queryable properties used as annotations in this
        # query.
        self._queryable_property_annotations = set()
        # A stack for queryable properties who are currently being annotated.
        # Required to correctly resolve dependencies and perform annotations.
        self._queryable_property_stack = []

    def _resolve_queryable_property(self, path):
        """
        Resolve the given path into a queryable property on the model
        associated with this query.

        :param collections.Sequence[str] path: The path to resolve (a string of
                                               Django's query expression split
                                               up by the lookup separator).
        :return: A 2-tuple containing a queryable property reference for the
                 resolved property and a list containing the parts of the path
                 that represent lookups (or transforms). The first item will be
                 None and the list will be empty if no queryable property could
                 be resolved.
        :rtype: (QueryablePropertyReference, list[str])
        """
        model = self.model
        property_ref, lookups = None, []
        # Try to follow the given path to allow to use queryable properties
        # across relations.
        for index, name in enumerate(path):
            try:
                related_model = get_related_model(model, name)
            except FieldDoesNotExist:
                try:
                    prop = get_queryable_property(model, name)
                except QueryablePropertyDoesNotExist:
                    # Neither a field nor a queryable property, so likely an
                    # invalid name. Do nothing and let Django deal with it.
                    pass
                else:
                    property_ref = QueryablePropertyReference(prop, model, tuple(path[:index]))
                    lookups = path[index + 1:]
                # The current name was not a field and either a queryable
                # property or invalid. Either way, resolving ends here.
                break
            else:
                if not related_model:
                    # A regular model field that doesn't represent a relation,
                    # meaning that no queryable property is involved.
                    break
                model = related_model
        return property_ref, lookups

    @contextmanager
    def _add_queryable_property_annotation(self, property_ref, full_group_by, select=False):
        """
        A context manager that adds a queryable property annotation to this
        query and performs management tasks around the annotation (stores the
        information if the queryable property annotation should be selected
        and populates the queryable property stack correctly). The context
        manager yields the actual resolved and applied annotation while the
        stack is still populated.

        :param QueryablePropertyReference property_ref: A reference containing
                                                        the queryable property
                                                        to annotate.
        :param bool full_group_by: Signals whether to use all fields of the
                                   query for the GROUP BY clause when dealing
                                   with an aggregate-based annotation or not.
        :param bool select: Signals whether the annotation should be selected
                            or not.
        """
        if property_ref in self._queryable_property_stack:
            raise QueryablePropertyError('Queryable property "{}" has a circular dependency and requires itself.'
                                         .format(property_ref.property))

        annotation_name = property_ref.full_path
        annotation_mask = set(self.annotations) if self.annotation_select_mask is None else self.annotation_select_mask
        self._queryable_property_stack.append(property_ref)
        try:
            if property_ref not in self._queryable_property_annotations:
                self.add_annotation(property_ref.get_annotation(), alias=annotation_name, is_summary=False)
                if not select:
                    self.set_annotation_mask(annotation_mask)
                self._queryable_property_annotations.add(property_ref)
            elif select and self.annotation_select_mask is not None:
                self.set_annotation_mask(annotation_mask.union((annotation_name,)))
            annotation = self.annotations[annotation_name]
            yield annotation
        finally:
            self._queryable_property_stack.pop()

        # Perform the required GROUP BY setup if the annotation contained
        # aggregates, which is normally done by QuerySet.annotate.
        if contains_aggregate(annotation):
            if full_group_by and not ANNOTATION_TO_AGGREGATE_ATTRIBUTES_MAP:
                # In recent Django versions, a full GROUP BY can be achieved by
                # simply setting group_by to True.
                self.group_by = True
            else:
                if full_group_by:  # pragma: no cover
                    # In old versions, the fields must be added to the selected
                    # fields manually and set_group_by must be called after.
                    opts = self.model._meta
                    self.add_fields([f.attname for f in getattr(opts, 'concrete_fields', opts.fields)], False)
                self.set_group_by()

    def _auto_annotate(self, path):
        """
        Try to resolve the given path into a queryable property and annotate
        the property as a non-selected property (if the property wasn't added
        as an annotation already). Do nothing if the path does not match a
        queryable property.

        :param collections.Sequence path: The path to resolve (a string of
                                          Django's query expression split up
                                          by the lookup separator).
        :return: The resolved annotation or None if the path couldn't be
                 resolved.
        """
        property_ref = self._resolve_queryable_property(path)[0]
        if not property_ref:
            return None
        full_group_by = bool(ANNOTATION_TO_AGGREGATE_ATTRIBUTES_MAP) and not self.select
        with self._add_queryable_property_annotation(property_ref, full_group_by) as annotation:
            return annotation

    def _postprocess_clone(self, clone):
        """
        Postprocess a query that was the result of cloning this query. This
        ensures that the cloned query also uses this mixin and that the
        queryable property attributes are initialized correctly.

        :param django.db.models.sql.Query clone: The cloned query.
        :return: The postprocessed cloned query.
        :rtype: django.db.models.sql.Query
        """
        if not isinstance(clone, QueryablePropertiesQueryMixin):
            QueryablePropertiesQueryMixin.inject_into_object(clone)
        else:
            clone.init_injected_attrs()
        clone._queryable_property_annotations.update(self._queryable_property_annotations)
        return clone

    def add_aggregate(self, aggregate, model=None, alias=None, is_summary=False):  # pragma: no cover
        # This method is called in older versions to add an aggregate, which
        # may be based on a queryable property annotation, which in turn must
        # be auto-annotated here.
        path = tuple(aggregate.lookup.split(LOOKUP_SEP))
        if self._queryable_property_stack:
            path = self._queryable_property_stack[-1].relation_path + path
        property_annotation = self._auto_annotate(path)
        if property_annotation:
            # If it is based on a queryable property annotation, annotating the
            # current aggregate cannot be delegated to Django as it couldn't
            # deal with annotations containing the lookup separator.
            aggregate.add_to_query(self, alias, LOOKUP_SEP.join(path), property_annotation, is_summary)
        else:
            # The overridden method also allows to set a default value for the
            # model parameter, which will be missing if add_annotation calls are
            # redirected to add_aggregate for older Django versions.
            model = model or self.model
            super(QueryablePropertiesQueryMixin, self).add_aggregate(aggregate, model, alias, is_summary)
        if self.annotation_select_mask is not None:
            self.set_annotation_mask(self.annotation_select_mask.union((alias,)))

    def add_filter(self, *args, **kwargs):  # pragma: no cover
        # The build_filter method was called add_filter in very old Django
        # versions. Since recent versions still have an add_filter method (for
        # different purposes), the queryable properties customizations should
        # only occur in old versions.
        if BUILD_FILTER_METHOD_NAME == 'add_filter':
            # Simply use the build_filter implementation that does all the
            # heavy lifting and is aware of the different methods in different
            # versions and therefore calls the correct super methods if
            # necessary.
            return self.build_filter(*args, **kwargs)
        return super(QueryablePropertiesQueryMixin, self).add_filter(*args, **kwargs)

    def add_ordering(self, *ordering, **kwargs):
        for field_name in ordering:
            # Ordering by a queryable property via simple string values
            # requires auto-annotating here, while a queryable property used
            # in a complex ordering expression is resolved through overridden
            # query methods.
            if isinstance(field_name, six.string_types) and field_name != '?':
                if field_name.startswith('-') or field_name.startswith('+'):
                    field_name = field_name[1:]
                self._auto_annotate(field_name.split(LOOKUP_SEP))
        return super(QueryablePropertiesQueryMixin, self).add_ordering(*ordering, **kwargs)

    def build_filter(self, filter_expr, *args, **kwargs):
        # Check if the given filter expression is meant to use a queryable
        # property. Therefore, the possibility of filter_expr not being of the
        # correct type must be taken into account (a case Django would cover
        # already, but the check for queryable properties MUST run first).
        try:
            arg, value = filter_expr
        except (TypeError, ValueError):
            # Invalid value - just treat it as "no queryable property found"
            # and delegate it to Django.
            property_ref = None
        else:
            property_ref, lookups = self._resolve_queryable_property(arg.split(LOOKUP_SEP))

        # If no queryable property could be determined for the filter
        # expression (either because a regular/non-existent field is referenced
        # or because the expression was a special or invalid value), call
        # Django's default implementation, which may in turn raise an
        # exception. Act the same way if the current top of the stack is used
        # to avoid infinite recursions.
        if not property_ref or (self._queryable_property_stack and self._queryable_property_stack[-1] == property_ref):
            # The base method has different names in different Django versions
            # (see comment on the constant definition).
            base_method = getattr(super(QueryablePropertiesQueryMixin, self), BUILD_FILTER_METHOD_NAME)
            return base_method(filter_expr, *args, **kwargs)

        q_obj = property_ref.get_filter(lookups, value)
        # Before applying the filter implemented by the property, check if
        # the property signals the need of its own annotation to function.
        # If so, add the annotation first to avoid endless recursion, since
        # resolved filter will likely contain the same property name again.
        context = nullcontext()
        if property_ref.property.filter_requires_annotation:
            full_group_by = bool(ANNOTATION_TO_AGGREGATE_ATTRIBUTES_MAP) and not self.select
            context = self._add_queryable_property_annotation(property_ref, full_group_by)

        with context:
            # Luckily, build_filter and _add_q use the same return value
            # structure, so an _add_q call can be used to actually create the
            # return value for the current call. The (_)add_q method has
            # different names in different Django versions (see comment on the
            # constant definition).
            method = getattr(self, ADD_Q_METHOD_NAME)
            return method(q_obj, **convert_build_filter_to_add_q_kwargs(**kwargs))

    def get_aggregation(self, *args, **kwargs):
        # If the query is to be used as a pure aggregate query (which might use
        # a subquery), all queryable property annotations must be added to the
        # select mask to avoid potentially empty SELECT clauses.
        if self.annotation_select_mask is not None and self._queryable_property_annotations:
            annotation_names = (property_ref.full_path for property_ref in self._queryable_property_annotations)
            self.set_annotation_mask(self.annotation_select_mask.union(annotation_names))
        return super(QueryablePropertiesQueryMixin, self).get_aggregation(*args, **kwargs)

    def need_force_having(self, q_object):  # pragma: no cover
        # Same as need_having, but for even older versions. Simply delegate to
        # need_having, which is aware of the different methods in different
        # versions and therefore calls the correct super methods if
        # necessary.
        return self.need_having(q_object)

    def need_having(self, obj):  # pragma: no cover
        # This method is used by older Django versions to figure out if the
        # filter represented by a Q object must be put in the HAVING clause of
        # the query. Since a queryable property might add an aggregate-based
        # annotation during the actual filter application, this method must
        # return True if a filter condition contains such a property.
        def is_aggregate_property(item, ignored_refs=set()):
            path = item[0].split(LOOKUP_SEP)
            property_ref, lookups = self._resolve_queryable_property(path)
            if not property_ref or property_ref in ignored_refs:
                return False
            if property_ref.property.filter_requires_annotation:
                if contains_aggregate(property_ref.get_annotation()):
                    return True
                ignored_refs = ignored_refs.union((property_ref,))
            # Also check the Q object returned by the property's get_filter
            # method as it may contain references to other properties that may
            # add aggregation-based annotations.
            predicate = partial(is_aggregate_property, ignored_refs=ignored_refs)
            return TreeNodeProcessor(property_ref.get_filter(lookups, item[1])).check_leaves(predicate)

        if isinstance(obj, Node) and TreeNodeProcessor(obj).check_leaves(is_aggregate_property):
            return True
        elif not isinstance(obj, Node) and is_aggregate_property(obj):
            return True
        # The base method has different names in different Django versions (see
        # comment on the constant definition).
        base_method = getattr(super(QueryablePropertiesQueryMixin, self), NEED_HAVING_METHOD_NAME)
        return base_method(obj)

    def resolve_ref(self, name, allow_joins=True, reuse=None, summarize=False, *args, **kwargs):
        # This method is used to resolve field names in complex expressions. If
        # a queryable property is used in such an expression, it needs to be
        # auto-annotated (while taking the stack into account) and returned.
        path = tuple(name.split(LOOKUP_SEP))
        if self._queryable_property_stack:
            path = self._queryable_property_stack[-1].relation_path + path
        property_annotation = self._auto_annotate(path)
        if property_annotation:
            if summarize:
                # Outer queries for aggregations need refs to annotations of
                # the inner queries.
                from django.db.models.expressions import Ref
                return Ref(name, property_annotation)
            return property_annotation
        return super(QueryablePropertiesQueryMixin, self).resolve_ref(name, allow_joins, reuse, summarize,
                                                                      *args, **kwargs)

    def setup_joins(self, names, *args, **kwargs):
        # This is a central method for resolving field names and joining the
        # required tables when dealing with paths that involve relations. To
        # also allow the usage of queryable properties across relations, the
        # relation path on top of the stack must be prepended to trick Django
        # into resolving correctly.
        if self._queryable_property_stack:
            names = self._queryable_property_stack[-1].relation_path + tuple(names)
        return super(QueryablePropertiesQueryMixin, self).setup_joins(names, *args, **kwargs)

    def clone(self, *args, **kwargs):
        obj = super(QueryablePropertiesQueryMixin, self).clone(*args, **kwargs)
        if QUERY_CHAIN_METHOD_NAME == 'clone':  # pragma: no cover
            obj = self._postprocess_clone(obj)
        return obj

    def chain(self, *args, **kwargs):
        obj = super(QueryablePropertiesQueryMixin, self).chain(*args, **kwargs)
        if QUERY_CHAIN_METHOD_NAME == 'chain':
            obj = self._postprocess_clone(obj)
        return obj



# encoding: utf-8

import pytest

from django import VERSION as DJANGO_VERSION

from ..models import (ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties,
                      CategoryWithClassBasedProperties, CategoryWithDecoratorBasedProperties,
                      VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties)

pytestmark = [pytest.mark.django_db, pytest.mark.usefixtures('versions')]


class TestAggregateAnnotations(object):

    @pytest.mark.parametrize('model, filters, expected_version_counts', [
        (ApplicationWithClassBasedProperties, {}, {3, 4}),
        (ApplicationWithClassBasedProperties, {'name__contains': 'cool'}, {3}),
        (ApplicationWithClassBasedProperties, {'version_count__gt': 3}, {4}),
        (ApplicationWithClassBasedProperties, {'version_count': 5}, set()),
        (ApplicationWithDecoratorBasedProperties, {}, {3, 4}),
        (ApplicationWithDecoratorBasedProperties, {'name__contains': 'cool'}, {3}),
        (ApplicationWithDecoratorBasedProperties, {'version_count__gt': 3}, {4}),
        (ApplicationWithDecoratorBasedProperties, {'version_count': 5}, set()),
    ])
    def test_values_to_limit_fields(self, model, filters, expected_version_counts):
        # Delete one version to create distinct version counts
        model.objects.get(name__contains='cool').versions.all()[0].delete()
        queryset = model.objects.filter(**filters).select_properties('version_count').values('version_count')
        version_counts = set(obj_dict['version_count'] for obj_dict in queryset)
        assert version_counts == expected_version_counts

    @pytest.mark.parametrize('model', [ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties])
    def test_values_for_custom_grouping(self, model):
        values = model.objects.values('common_data').select_properties('version_count')
        assert len(values) == 1
        assert values[0]['version_count'] == 8

    @pytest.mark.parametrize('model, values, property_name, filter_value, expected_count', [
        (ApplicationWithClassBasedProperties, ('common_data',), 'version_count', 7, 1),
        (ApplicationWithDecoratorBasedProperties, ('common_data',), 'version_count', 7, 1),
        (CategoryWithClassBasedProperties, ('pk', 'applications__pk'), 'applications__version_count', 4, 2),
        (CategoryWithDecoratorBasedProperties, ('pk', 'applications__pk'), 'applications__version_count', 4, 2),
    ])
    def test_values_without_property_selection(self, model, values, property_name, filter_value, expected_count):
        version_model = (VersionWithClassBasedProperties if 'ClassBased' in model.__name__
                         else VersionWithDecoratorBasedProperties)
        # Delete one version to create distinct version counts
        version_model.objects.get(application__name__contains='cool', major=2).delete()
        queryset = model.objects.values(*values).filter(**{property_name: filter_value})
        assert len(queryset) == expected_count
        assert len(set(tuple(obj_dict.items()) for obj_dict in queryset)) == expected_count
        assert all(property_name not in obj_dict for obj_dict in queryset)

    @pytest.mark.parametrize('model, values, property_name, expected_count', [
        (ApplicationWithClassBasedProperties, ('common_data',), 'version_count', 1),
        (ApplicationWithDecoratorBasedProperties, ('common_data',), 'version_count', 1),
        (CategoryWithClassBasedProperties, ('pk', 'applications__pk'), 'applications__version_count', 3),
        (CategoryWithDecoratorBasedProperties, ('pk', 'applications__pk'), 'applications__version_count', 3),
    ])
    def test_values_with_order_by_property(self, model, values, property_name, expected_count):
        # In Django versions below 1.8, annotations used for ordering MUST be
        # selected, which expectedly tinkers with the GROUPING.
        expected_property_selection = DJANGO_VERSION < (1, 8)
        expected_queryset_len = expected_count + int(expected_property_selection and property_name == 'version_count')
        queryset = model.objects.order_by(property_name).values(*values)
        assert len(queryset) == expected_queryset_len
        assert len(set(tuple(obj_dict.items()) for obj_dict in queryset)) == expected_count
        assert all((property_name in obj_dict) is expected_property_selection for obj_dict in queryset)

    @pytest.mark.parametrize('model', [ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties])
    def test_distinct_property_values(self, model):
        queryset = model.objects.select_properties('version_count').values('version_count').distinct()
        assert queryset.count() == len(queryset) == 1
        assert queryset[0]['version_count'] == 4

    @pytest.mark.parametrize('model, filters, expected_version_counts', [
        (ApplicationWithClassBasedProperties, {}, {3, 4}),
        (ApplicationWithClassBasedProperties, {'name__contains': 'cool'}, {3}),
        (ApplicationWithClassBasedProperties, {'version_count__gt': 3}, {4}),
        (ApplicationWithClassBasedProperties, {'version_count': 5}, set()),
        (ApplicationWithDecoratorBasedProperties, {}, {3, 4}),
        (ApplicationWithDecoratorBasedProperties, {'name__contains': 'cool'}, {3}),
        (ApplicationWithDecoratorBasedProperties, {'version_count__gt': 3}, {4}),
        (ApplicationWithDecoratorBasedProperties, {'version_count': 5}, set()),
    ])
    def test_values_list(self, model, filters, expected_version_counts):
        # Delete one version to create distinct version counts
        model.objects.get(name__contains='cool').versions.all()[0].delete()
        queryset = model.objects.filter(**filters).select_properties('version_count').values_list('version_count',
                                                                                                  flat=True)
        assert set(queryset) == expected_version_counts


@pytest.mark.skipif(DJANGO_VERSION < (1, 8), reason="Expression-based annotations didn't exist before Django 1.8")
class TestExpressionAnnotations(object):

    @pytest.mark.parametrize('model, filters, expected_versions', [
        (VersionWithClassBasedProperties, {}, {'1.2.3', '1.3.0', '1.3.1', '2.0.0'}),
        (VersionWithClassBasedProperties, {'patch': 0}, {'1.3.0', '2.0.0'}),
        (VersionWithClassBasedProperties, {'major_minor': '1.3'}, {'1.3.0', '1.3.1'}),
        (VersionWithClassBasedProperties, {'version': '2.0.0'}, {'2.0.0'}),
        (VersionWithDecoratorBasedProperties, {}, {'1.2.3', '1.3.0', '1.3.1', '2.0.0'}),
        (VersionWithDecoratorBasedProperties, {'patch': 0}, {'1.3.0', '2.0.0'}),
        (VersionWithDecoratorBasedProperties, {'major_minor': '1.3'}, {'1.3.0', '1.3.1'}),
        (VersionWithDecoratorBasedProperties, {'version': '2.0.0'}, {'2.0.0'}),
    ])
    def test_values_to_limit_fields(self, model, filters, expected_versions):
        queryset = model.objects.filter(**filters).select_properties('version').values('version')
        versions = set(obj_dict['version'] for obj_dict in queryset)
        assert versions == expected_versions

    @pytest.mark.skipif(DJANGO_VERSION < (1, 11), reason="Explicit subqueries didn't exist before Django 1.11")
    @pytest.mark.parametrize('model', [ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties])
    def test_group_by_property(self, model):
        queryset = model.objects.select_properties('highest_version').values('highest_version')
        queryset = queryset.select_properties('version_count')
        assert queryset.count() == len(queryset) == 1
        assert queryset[0]['version_count'] == 8

    @pytest.mark.parametrize('model, values, property_name, filter_value, expected_count', [
        (VersionWithClassBasedProperties, ('major',), 'changes_or_default', '(No data)', 6),
        (VersionWithDecoratorBasedProperties, ('major',), 'changes_or_default', '(No data)', 6),
        (ApplicationWithClassBasedProperties, ('name',), 'versions__changes_or_default', '(No data)', 6),
        (ApplicationWithDecoratorBasedProperties, ('name',), 'versions__changes_or_default', '(No data)', 6),
    ])
    def test_values_without_property_selection(self, model, values, property_name, filter_value, expected_count):
        queryset = model.objects.values(*values).filter(**{property_name: filter_value})
        assert len(queryset) == expected_count
        assert all(property_name not in obj_dict for obj_dict in queryset)

    @pytest.mark.parametrize('model', [VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties])
    def test_distinct_property_values(self, model):
        queryset = model.objects.select_properties('version').values('version').distinct()
        assert queryset.count() == len(queryset) == 4
        assert set(obj_dict['version'] for obj_dict in queryset) == {'1.2.3', '1.3.0', '1.3.1', '2.0.0'}

    @pytest.mark.parametrize('model, filters, expected_versions', [
        (VersionWithClassBasedProperties, {}, {'1.2.3', '1.3.0', '1.3.1', '2.0.0'}),
        (VersionWithClassBasedProperties, {'patch': 0}, {'1.3.0', '2.0.0'}),
        (VersionWithClassBasedProperties, {'major_minor': '1.3'}, {'1.3.0', '1.3.1'}),
        (VersionWithClassBasedProperties, {'version': '2.0.0'}, {'2.0.0'}),
        (VersionWithDecoratorBasedProperties, {}, {'1.2.3', '1.3.0', '1.3.1', '2.0.0'}),
        (VersionWithDecoratorBasedProperties, {'patch': 0}, {'1.3.0', '2.0.0'}),
        (VersionWithDecoratorBasedProperties, {'major_minor': '1.3'}, {'1.3.0', '1.3.1'}),
        (VersionWithDecoratorBasedProperties, {'version': '2.0.0'}, {'2.0.0'}),
    ])
    def test_values_list(self, model, filters, expected_versions):
        queryset = model.objects.filter(**filters).select_properties('version').values_list('version', flat=True)
        assert set(queryset) == expected_versions



# encoding: utf-8

from __future__ import unicode_literals

from django.core.exceptions import FieldError
try:  # pragma: no cover
    from django.core.exceptions import FieldDoesNotExist
except ImportError:  # pragma: no cover
    from django.db.models.fields import FieldDoesNotExist


class QueryablePropertyError(FieldError):
    """Some kind of problem with a queryable property."""


class QueryablePropertyDoesNotExist(FieldDoesNotExist):
    """The requested queryable property does not exist."""



# encoding: utf-8

SECRET_KEY = '1337' * 8

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'queryable_properties.db',
        'USER': '',
        'PASSWORD': '',
        'HOST': '',
        'PORT': '',
    }
}


INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.admin',
    'django.contrib.admindocs',
    'queryable_properties',
    'tests',
)

SITE_ID = 1



# encoding: utf-8

from .base import QueryableProperty, queryable_property
from .cache_behavior import CACHE_RETURN_VALUE, CACHE_VALUE, CLEAR_CACHE, DO_NOTHING
from .common import AggregateProperty, RangeCheckProperty, ValueCheckProperty
from .mixins import AnnotationMixin, boolean_filter, LookupFilterMixin, lookup_filter, SetterMixin, UpdateMixin

__all__ = (
    'QueryableProperty', 'queryable_property',
    'CACHE_RETURN_VALUE', 'CACHE_VALUE', 'CLEAR_CACHE', 'DO_NOTHING',
    'AggregateProperty', 'RangeCheckProperty', 'ValueCheckProperty',
    'AnnotationMixin', 'boolean_filter', 'LookupFilterMixin', 'lookup_filter', 'SetterMixin', 'UpdateMixin',
)



# encoding: utf-8

import pytest

from django import VERSION as DJANGO_VERSION
from django.db import models

from ..conftest import Concat, Value
from ..models import (ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties,
                      VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties)

pytestmark = [pytest.mark.django_db, pytest.mark.usefixtures('versions')]


class TestAggregateAnnotations(object):

    @pytest.mark.parametrize('model, filters', [
        (ApplicationWithClassBasedProperties, {}),
        (ApplicationWithClassBasedProperties, {'version_count__gt': 3}),
        (ApplicationWithDecoratorBasedProperties, {}),
        (ApplicationWithDecoratorBasedProperties, {'version_count__gt': 3}),
    ])
    def test_cached_annotation_value(self, model, filters):
        # Filter both before and after the select_properties call to check if
        # the annotation gets selected correctly regardless
        queryset = model.objects.filter(**filters).select_properties('version_count', 'major_sum').filter(**filters)
        assert 'version_count' in queryset.query.annotations
        assert 'major_sum' in queryset.query.annotations
        assert all(model.version_count._has_cached_value(obj) for obj in queryset)
        assert all(model.major_sum._has_cached_value(obj) for obj in queryset)

    @pytest.mark.parametrize('model, limit, expected_total', [
        (ApplicationWithClassBasedProperties, None, 8),
        (ApplicationWithClassBasedProperties, 1, 4),
        (ApplicationWithDecoratorBasedProperties, None, 8),
        (ApplicationWithDecoratorBasedProperties, 1, 4),
    ])
    def test_aggregate_based_on_queryable_property(self, model, limit, expected_total):
        result = model.objects.all()[:limit].aggregate(total_version_count=models.Sum('version_count'))
        assert result['total_version_count'] == expected_total

    @pytest.mark.parametrize('model, limit, expected_total', [
        (VersionWithClassBasedProperties, None, 32),
        (VersionWithClassBasedProperties, 4, 16),
        (VersionWithDecoratorBasedProperties, None, 32),
        (VersionWithDecoratorBasedProperties, 4, 16),
    ])
    def test_aggregate_based_on_queryable_property_across_relation(self, model, limit, expected_total):
        result = model.objects.all()[:limit].aggregate(total_version_count=models.Sum('application__version_count'))
        assert result['total_version_count'] == expected_total

    @pytest.mark.skipif(DJANGO_VERSION < (1, 8), reason="Expression-based annotations didn't exist before Django 1.8")
    @pytest.mark.parametrize('model, annotation', [
        (VersionWithClassBasedProperties, models.F('application__version_count')),
        (VersionWithDecoratorBasedProperties, models.F('application__version_count')),
    ])
    def test_annotation_based_on_queryable_property_across_relation(self, model, annotation):
        model.objects.all()[0].delete()
        queryset = model.objects.annotate(annotation=annotation)
        assert all(obj.annotation == obj.application.version_count for obj in queryset)

    @pytest.mark.parametrize('model', [ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties])
    def test_iterator(self, model):
        queryset = model.objects.filter(version_count=4).select_properties('version_count')
        for application in queryset.iterator():
            assert model.version_count._has_cached_value(application)
            assert application.version_count == 4
        assert queryset._result_cache is None

    @pytest.mark.parametrize('model', [ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties])
    def test_removed_annotation(self, model):
        """
        Test that queries can still be performed even if queryable property annotations have been manually removed from
        the queryset.
        """
        queryset = model.objects.select_properties('version_count')
        del queryset.query.annotations['version_count']
        assert bool(queryset)
        assert all(not model.version_count._has_cached_value(obj) for obj in queryset)


@pytest.mark.skipif(DJANGO_VERSION < (1, 8), reason="Expression-based annotations didn't exist before Django 1.8")
class TestExpressionAnnotations(object):

    @pytest.mark.parametrize('model, filters', [
        (VersionWithClassBasedProperties, {}),
        (VersionWithDecoratorBasedProperties, {}),
        (VersionWithClassBasedProperties, {'version': '1.2.3'}),
        (VersionWithDecoratorBasedProperties, {'version': '1.2.3'}),
    ])
    def test_cached_annotation_value(self, model, filters):
        # Filter both before and after the select_properties call to check if
        # the annotation gets selected correctly regardless
        queryset = model.objects.filter(**filters).select_properties('version').filter(**filters)
        assert 'version' in queryset.query.annotations
        assert all(model.version._has_cached_value(obj) for obj in queryset)

    @pytest.mark.parametrize('model, property_name, annotation, expected_count, record_checker', [
        (VersionWithClassBasedProperties, 'version', models.F('version'), 8,
         lambda obj: obj.annotation == obj.version),
        (VersionWithDecoratorBasedProperties, 'version', models.F('version'), 8,
         lambda obj: obj.annotation == obj.version),
        (VersionWithClassBasedProperties, 'version', Concat(Value('V'), 'version'), 8,
         lambda obj: obj.annotation == 'V' + obj.version),
        (VersionWithDecoratorBasedProperties, 'version', Concat(Value('V'), 'version'), 8,
         lambda obj: obj.annotation == 'V' + obj.version),
        (ApplicationWithClassBasedProperties, 'versions__version', models.F('versions__version'), 8,
         lambda obj: obj.annotation in ('1.2.3', '1.3.0', '1.3.1', '2.0.0')),
        (ApplicationWithDecoratorBasedProperties, 'versions__version', models.F('versions__version'), 8,
         lambda obj: obj.annotation in ('1.2.3', '1.3.0', '1.3.1', '2.0.0')),
        (ApplicationWithClassBasedProperties, 'versions__version', Concat(Value('V'), 'versions__version'), 8,
         lambda obj: obj.annotation in ('V1.2.3', 'V1.3.0', 'V1.3.1', 'V2.0.0')),
        (ApplicationWithDecoratorBasedProperties, 'versions__version', Concat(Value('V'), 'versions__version'), 8,
         lambda obj: obj.annotation in ('V1.2.3', 'V1.3.0', 'V1.3.1', 'V2.0.0')),
    ])
    def test_annotation_based_on_queryable_property(self, model, property_name, annotation, expected_count,
                                                    record_checker):
        queryset = model.objects.annotate(annotation=annotation)
        assert queryset.count() == len(queryset) == expected_count
        assert all(record_checker(obj) for obj in queryset)
        if '__' not in property_name:
            # Check that a property annotation used implicitly by another
            # annotation does not lead to a selection of the property
            # annotation
            prop = getattr(model, property_name)
            assert all(not prop._has_cached_value(obj) for obj in queryset)

    @pytest.mark.parametrize('model', [VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties])
    def test_iterator(self, model):
        queryset = model.objects.filter(major_minor='2.0').select_properties('version')
        for version in queryset.iterator():
            assert model.version._has_cached_value(version)
            assert version.version == '2.0.0'
        assert queryset._result_cache is None



# encoding: utf-8


def CLEAR_CACHE(prop, obj, value, return_value):
    """
    Setter cache behavior function that will clear the cached value for a
    cached queryable property on objects after the setter was used.

    :param QueryableProperty prop: The property whose setter was used.
    :param django.db.models.Model obj: The object the setter was used on.
    :param value: The value that was passed to the setter.
    :param return_value: The return value of the setter function/method.
    """
    prop._clear_cached_value(obj)


def CACHE_VALUE(prop, obj, value, return_value):
    """
    Setter cache behavior function that will update the cache for the cached
    queryable property on the object in question with the (raw) value that was
    passed to the setter.

    :param QueryableProperty prop: The property whose setter was used.
    :param django.db.models.Model obj: The object the setter was used on.
    :param value: The value that was passed to the setter.
    :param return_value: The return value of the setter function/method.
    """
    prop._set_cached_value(obj, value)


def CACHE_RETURN_VALUE(prop, obj, value, return_value):
    """
    Setter cache behavior function that will update the cache for the cached
    queryable property on the object in question with the return value of the
    setter function/method.

    :param QueryableProperty prop: The property whose setter was used.
    :param django.db.models.Model obj: The object the setter was used on.
    :param value: The value that was passed to the setter.
    :param return_value: The return value of the setter function/method.
    """
    prop._set_cached_value(obj, return_value)


def DO_NOTHING(prop, obj, value, return_value):
    """
    Setter cache behavior function that will do nothing after the setter of
    a cached queryable property was used, retaining previously cached values.

    :param QueryableProperty prop: The property whose setter was used.
    :param django.db.models.Model obj: The object the setter was used on.
    :param value: The value that was passed to the setter.
    :param return_value: The return value of the setter function/method.
    """
    pass



#!/usr/bin/env python
# encoding: utf-8

from __future__ import unicode_literals

from distutils.cmd import Command
import io
import os
from setuptools import find_packages, setup
import subprocess

HERE = os.path.abspath(os.path.dirname(__file__))


def read_file(*path_parts, **kwargs):
    with io.open(os.path.join(HERE, *path_parts), **kwargs) as f:
        return f.read()


META = {}
exec(read_file('queryable_properties', '__init__.py', mode='rb'), {}, META)


class PytestCommand(Command):
    """
    Distutils command to run tests via pytest in the current environment.
    """
    description = 'Run tests in the current environment using pytest.'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import py
        return py.test.cmdline.main([])


class ToxCommand(Command):
    """
    Distutils command to run tests via tox, which must be installed and be able
    to access all supported python versions.
    """
    description = ('Run tests using tox, which must already be installed and be able to access all supported python '
                   'versions to create its environments.')
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        return subprocess.call(['tox'])


setup(
    name='django-queryable-properties',
    version=META['__version__'],
    description=META['__doc__'],
    long_description='\n\n'.join((read_file('README.rst', encoding='utf-8'),
                                  read_file('CHANGELOG.rst', encoding='utf-8'))),
    author=META['__author__'],
    author_email=META['__email__'],
    maintainer=META['__maintainer__'],
    maintainer_email=META['__email__'],
    url='https://github.com/W1ldPo1nter/django-queryable-properties',
    license='BSD',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.4',
        'Framework :: Django :: 1.5',
        'Framework :: Django :: 1.6',
        'Framework :: Django :: 1.7',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet',
    ],
    packages=find_packages(include='queryable_properties*'),
    include_package_data=True,
    install_requires=[
        'Django>=1.4',
        'six>=1.9',
    ],
    tests_require=[
        'Django>=1.4',
        'six>=1.9',
        'coverage',
        'flake8',
        'mock',
        'pydocstyle',
        'pytest',
        'pytest-cov',
        'pytest-django',
        'pytest-pep8',
        'pytest-pythonpath',
    ],
    cmdclass={
        'test': PytestCommand,
        'tox': ToxCommand,
    }
)
